from aerosandbox import *

a = Airplane(
    name="Single Wing",
    xyz_ref=[0, 0, 0],
    wings=[
        Wing(
            name="Wing",
            xyz_le=[0, 0, 0],
            symmetric=True,
            xsecs=[
                WingXSec(
                    xyz_le=[0, 0, 0],
                    chord=0.5,
                    twist=0,
                    airfoil=Airfoil(name="naca4412")
                ),
                WingXSec(
                    xyz_le=[0, 1, 0],
                    chord=0.5,
                    twist=0,
                    airfoil=Airfoil(name="naca4412")
                )
            ]
        )
    ]
)
a.set_ref_dims_from_wing()

a.set_paneling_everywhere(n_chordwise_panels=30, n_spanwise_panels=30)

ap = panel1(
    airplane=a,
    op_point=OperatingPoint(velocity=10,
                            alpha=15,
                            beta=0),
)
ap.run()
ap.draw(
    shading_type="doublet_strengths",
    streamlines_type=np.vstack((
        linspace_3D((-0.05, -0.95, -0.15), (-0.05, -0.95, 0.03), 30),
        linspace_3D((-0.05, 0, -0.15), (-0.05, 0, 0.03), 30),
    )),
    points_type=None
)

# Answer you should get: (XFLR5)
# CL = 0.320
# CDi = 0.008
# CL/CDi = 40.157
# Cm = -0.074



from aerosandbox.tools.airfoil_fitter.airfoil_fitter import *

if __name__ == '__main__':

    # a = Airfoil(name="HALE_03 (Thiqboi)", coordinates="C:/Projects/Github/Airfoils/HALE_03.dat")
    a = Airfoil("e216")
    # a = Airfoil("naca0008")

    try:
        with open("%s.pkl" % a.name, "rb") as f:
            af = pickle.load(f)
    except:
        a.get_xfoil_data()
        af = AirfoilFitter(a)

        with open("%s.pkl" % a.name, "wb+") as f:
            pickle.dump(af, f)

    # af.airfoil.plot_xfoil_data_contours()
    # af.airfoil.plot_xfoil_data_polars()
    # af.plot_xfoil_alpha_Re('Cl')
    # af.plot_xfoil_alpha_Re('Cd', log_z=True)
    Cl = af.fit_xfoil_data_Cl(plot_fit=True)
    Cd = af.fit_xfoil_data_Cd(plot_fit=True)

    # Get data
    y_data = af.airfoil.xfoil_data_1D['Cl'] / af.airfoil.xfoil_data_1D['Cd']

    # Get model
    n = 60
    linspace = lambda x: np.linspace(np.min(x), np.max(x), n)
    logspace = lambda x: np.logspace(np.log10(np.min(x)), np.log10(np.max(x)), n)
    x1 = linspace(af.airfoil.xfoil_data_1D['alpha'])
    x2 = logspace(af.airfoil.xfoil_data_1D['Re'])
    X1, X2 = np.meshgrid(x1, x2)
    y_model = np.array(Cl(X1, X2) / Cd(X1, X2))

    fig = go.Figure()
    # fig.add_trace(
    #     go.Scatter3d(
    #         x=af.airfoil.xfoil_data_1D['alpha'],
    #         y=af.airfoil.xfoil_data_1D['Re'],
    #         z=af.airfoil.xfoil_data_1D['Cl']/af.airfoil.xfoil_data_1D['Cd'],
    #         mode="markers",
    #         marker=dict(
    #             size=2,
    #             color="black"
    #         )
    #     )
    # )
    # fig.add_trace(
    #     go.Surface(
    #         contours={
    #             # "x": {"show": True, "start": -20, "end": 20, "size": 1},
    #             # "y": {"show": True, "start": 1e4, "end": 1e6, "size": 1e5},
    #             # "z": {"show": True, "start": -5, "end": 5, "size": 0.1}
    #         },
    #         x=x1,
    #         y=x2,
    #         z=y_model,
    #         surfacecolor=y_model,
    #         colorscale="plasma",
    #         # flatshading=True
    #     )
    # )
    # fig.update_layout(
    #     scene=dict(
    #         xaxis=dict(
    #             title="Alpha"
    #         ),
    #         yaxis=dict(
    #             type='log',
    #             title="Re"
    #         ),
    #         zaxis=dict(
    #             type='linear',
    #             title="L/D"
    #         ),
    #     ),
    #     title="L/D"
    # )
    fig.add_trace(
        go.Mesh3d(
            x=af.airfoil.xfoil_data_1D["alpha"],
            y=af.airfoil.xfoil_data_1D["Re"],
            z=np.array(
                (Cl(af.airfoil.xfoil_data_1D["alpha"], af.airfoil.xfoil_data_1D["Re"]) / Cd(af.airfoil.xfoil_data_1D["alpha"], af.airfoil.xfoil_data_1D["Re"])) /
                (af.airfoil.xfoil_data_1D["Cl"] / af.airfoil.xfoil_data_1D["Cd"])
            ).reshape(-1),
        )
    )
    fig.update_layout(
        scene=dict(
            xaxis=dict(
                title="Alpha"
            ),
            yaxis=dict(
                type='log',
                title="Re"
            ),
            zaxis=dict(
                type='linear',
                title="L/D",
                range=(0,2)
            ),
        ),
        title="L/D"
    )
    fig.show()

    # with open("func.pkl", "wb+") as f:
    #     pickle.dump(func, f)
    # print(
    #     func(0, 1e6)
    # )



from aerosandbox_legacy_v0 import *

paper_airplane = Airplane(
    name="Conventional",
    xyz_ref=[0, 0, 0], # CG location
    wings=[
        Wing(
            name="Main Wing",
            xyz_le=[0, 0, 0], # Coordinates of the wing's leading edge
            symmetric=True,
            xsecs=[ # The wing's cross ("X") sections
                WingXSec(  # Root
                    xyz_le=[0, 0, 0], # Coordinates of the XSec's leading edge, relative to the wing's leading edge.
                    chord=0.0254*11,
                    twist=0, # degrees
                    airfoil=Airfoil(name="naca0003"),
                    control_surface_type='symmetric',  # Flap # Control surfaces are applied between a given XSec and the next one.
                    control_surface_deflection=0, # degrees
                    control_surface_hinge_point=0.75 # as chord fraction
                ),
                WingXSec(  # Mid
                    xyz_le=[0.0254*1.349, 0.0254*0.25, 0.0254*0.5],
                    chord=0.0254*9.651,
                    twist=0,
                    airfoil=Airfoil(name="naca0003"),
                    control_surface_type='symmetric',  # Aileron
                    control_surface_deflection=0,
                    control_surface_hinge_point=0.75
                ),
                WingXSec(  # Tip
                    xyz_le=[0.0254*10.260, 0.0254*3.907, 0.0254*1],
                    chord=0.0254*0.740,
                    twist=0,
                    airfoil=Airfoil(name="naca0003"),
                )
            ]
        )
    ]
)
paper_airplane.set_paneling_everywhere(20, 20)

ap = vlm3(
    airplane=paper_airplane,
    op_point=OperatingPoint(
        velocity=10,
        alpha=5,
        beta=0,
        p=0,
        q=0,
        r=0,
    ),
)
ap.run()
ap.draw()

# Answer you should get: (XFLR5)
# CL = 0.797
# CDi = 0.017
# CL/CDi = 47.211
# Cm = -0.184



from aerosandbox.geometry.airfoil import *
import pytest

@pytest.fixture
def naca4412():
    return Airfoil("naca4412")

@pytest.fixture
def e216():
    return Airfoil("e216")

def test_TE_angle(naca4412):
    assert naca4412.TE_angle() == pytest.approx(14.766578406372423)

def test_local_thickness(e216):
    assert e216.local_thickness(0.5) == pytest.approx(0.08730287761717835)

def LE_index_test(e216):
    assert e216.LE_index() == 32

def repanel_test(naca4412):
    naca4412.repanel(n_points_per_side=200, inplace=True)
    # assert

if __name__ == '__main__':
    pytest.main()


import casadi as cas


def propeller_shaft_power_from_thrust(
        thrust_force,
        area_propulsive,
        airspeed,
        rho,
        propeller_coefficient_of_performance=0.8,
):
    """
    Using dynamic disc actuator theory, gives the shaft power required to generate
    a certain amount of thrust.

    Source: https://web.mit.edu/16.unified/www/FALL/thermodynamics/notes/node86.html

    :param thrust_force: Thrust force [N]
    :param area_propulsive: Total disc area of all propulsive surfaces [m^2]
    :param airspeed: Airspeed [m/s]
    :param rho: Air density [kg/m^3]
    :param propeller_coefficient_of_performance: propeller coeff. of performance (due to viscous losses) [unitless]
    :return: Shaft power [W]
    """
    return 0.5 * thrust_force * airspeed * (
            cas.sqrt(
                thrust_force / (area_propulsive * airspeed ** 2 * rho / 2) + 1
            ) + 1
    ) / propeller_coefficient_of_performance


def mass_hpa_propeller(
        diameter,
        max_power,
        include_variable_pitch_mechanism=False
):
    """
    Returns the estimated mass of a propeller assembly for low-disc-loading applications (human powered airplane, paramotor, etc.)

    :param diameter: diameter of the propeller [m]
    :param max_power: maximum power of the propeller [W]
    :param include_variable_pitch_mechanism: boolean, does this propeller have a variable pitch mechanism?
    :return: estimated weight [kg]
    """
    smoothmax = lambda value1, value2, hardness: cas.log(
        cas.exp(hardness * value1) + cas.exp(hardness * value2)) / hardness  # soft maximum

    mass_propeller = (
            0.495 *
            (diameter / 1.25) ** 1.6 *
            smoothmax(0.6, max_power / 14914, hardness=5) ** 2
    )  # Baselining to a 125cm E-Props Top 80 Propeller for paramotor, with some sketchy scaling assumptions
    # Parameters on diameter exponent and min power were chosen such that Daedalus propeller is roughly on the curve.

    mass_variable_pitch_mech = 216.8 / 800 * mass_propeller
    # correlation to Daedalus data: http://journals.sfu.ca/ts/index.php/ts/article/viewFile/760/718
    if include_variable_pitch_mechanism:
        mass_propeller += mass_variable_pitch_mech

    return mass_propeller


def mass_gearbox(
        power,
        rpm_in,
        rpm_out,
):
    """
    Estimates the mass of a gearbox.

    Based on data from NASA/TM-2009-215680, available here:
        https://ntrs.nasa.gov/archive/nasa/casi.ntrs.nasa.gov/20090042817.pdf

    R^2 = 0.92 to the data.

    To quote this document:
        "The correlation was developed based on actual weight
        data from over fifty rotorcrafts, tiltrotors, and turboprop
        aircraft."

    Data fits in the NASA document were thrown out and refitted to extrapolate more sensibly; see:
        C:\Projects\GitHub\AeroSandbox\studies\GearboxMassFits
    :param power: Shaft power through the gearbox [W]
    :param rpm_in: RPM of the input to the gearbox [rpm]
    :param rpm_out: RPM of the output of the gearbox [rpm]
    :return: Estimated mass of the gearbox [kg]
    """
    power_hp = power / 745.7

    beta = (power_hp / rpm_out) ** 0.75 * (rpm_in / rpm_out) ** 0.15
    # Beta is a parametric value that tends to collapse the masses of gearboxes onto a line.
    # Data fit is considered tightly valid for gearboxes with 1 < beta < 100. Sensible extrapolations are made beyond that.

    p1 = 1.0445171124733774
    p2 = 2.0083615496306910

    mass_lb = 10 ** (p1 * cas.log10(beta) + p2)

    mass = mass_lb / 2.20462262185

    return mass


if __name__ == '__main__':
    import matplotlib.style as style

    style.use("seaborn")

    # Daedalus propeller
    print(
        mass_hpa_propeller(
            diameter=3.4442,
            max_power=177.93 * 8.2,  # max thrust at cruise speed
            include_variable_pitch_mechanism=False
        )
    )  # Should weight ca. 800 grams

    print(mass_gearbox(
        power=3000,
        rpm_in=6000,
        rpm_out=600
    ))



import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
import numpy as np
import casadi as cas
import matplotlib.pyplot as plt
import matplotlib.style as style
import seaborn as sns
plt.ion()
sns.set(font_scale=1)

# Set the rendering to happen in browser
pio.renderers.default = "browser"

def reflect_over_XZ_plane(input_vector):
    # Takes in a vector or an array and flips the y-coordinates.
    output_vector = input_vector
    shape = output_vector.shape
    if len(shape) == 1 and shape[0] == 3:
        output_vector = output_vector * cas.vertcat(1, -1, 1)
    elif len(shape) == 2 and shape[1] == 1 and shape[0] == 3:  # Vector of 3 items
        output_vector = output_vector * cas.vertcat(1, -1, 1)
    elif len(shape) == 2 and shape[1] == 3:  # 2D Nx3 vector
        output_vector = cas.horzcat(output_vector[:, 0], -1 * output_vector[:, 1], output_vector[:, 2])
    # elif len(shape) == 3 and shape[2] == 3:  # 3D MxNx3 vector
    #     output_vector = output_vector * cas.array([1, -1, 1])
    else:
        raise Exception("Invalid input for reflect_over_XZ_plane!")

    return output_vector


class Figure3D:
    def __init__(self):
        self.fig = go.Figure()

        # Vertices of the faces
        self.x_face = []
        self.y_face = []
        self.z_face = []

        # Connectivity and color of the faces
        self.i_face = []
        self.j_face = []
        self.k_face = []
        self.intensity_face = []

        # Vertices of the lines
        self.x_line = []
        self.y_line = []
        self.z_line = []

        # Vertices of the streamlines
        self.x_streamline = []
        self.y_streamline = []
        self.z_streamline = []

    def add_line(self,
                 points,
                 mirror=False,
                 ):
        """
        Adds a line (or series of lines) to draw.
        :param points: an iterable with an arbitrary number of items. Each item is a 3D point, represented as an iterable of length 3.
        :param mirror: Should we also draw a version that's mirrored over the XZ plane? [boolean]
        :return: None

        E.g. add_line([(0, 0, 0), (1, 0, 0)])
        """
        for p in points:
            self.x_line.append(float(p[0]))
            self.y_line.append(float(p[1]))
            self.z_line.append(float(p[2]))
        self.x_line.append(None)
        self.y_line.append(None)
        self.z_line.append(None)
        if mirror:
            reflected_points = [reflect_over_XZ_plane(point) for point in points]
            self.add_line(
                points=reflected_points,
                mirror=False
            )

    def add_streamline(self,
                       points,
                       mirror=False,
                       ):
        """
        Adds a line (or series of lines) to draw.
        :param points: an iterable with an arbitrary number of items. Each item is a 3D point, represented as an iterable of length 3.
        :param mirror: Should we also draw a version that's mirrored over the XZ plane? [boolean]
        :return: None

        E.g. add_line([(0, 0, 0), (1, 0, 0)])
        """
        for p in points:
            self.x_streamline.append(float(p[0]))
            self.y_streamline.append(float(p[1]))
            self.z_streamline.append(float(p[2]))
        self.x_streamline.append(None)
        self.y_streamline.append(None)
        self.z_streamline.append(None)
        if mirror:
            reflected_points = [reflect_over_XZ_plane(point) for point in points]
            self.add_streamline(
                points=reflected_points,
                mirror=False
            )

    def add_tri(self,
                points,
                intensity=0,
                outline=False,
                mirror=False,
                ):
        """
        Adds a triangular face to draw.
        :param points: an iterable with 3 items. Each item is a 3D point, represented as an iterable of length 3.
        :param intensity: Intensity associated with this face
        :param outline: Do you want to outline this triangle? [boolean]
        :param mirror: Should we also draw a version that's mirrored over the XZ plane? [boolean]
        :return: None

        E.g. add_face([(0, 0, 0), (1, 0, 0), (0, 1, 0)])
        """
        if not len(points) == 3:
            raise ValueError("'points' must have exactly 3 items!")
        for p in points:
            self.x_face.append(float(p[0]))
            self.y_face.append(float(p[1]))
            self.z_face.append(float(p[2]))
            self.intensity_face.append(intensity)
        indices_added = np.arange(len(self.x_face) - 3, len(self.x_face))
        self.i_face.append(indices_added[0])
        self.j_face.append(indices_added[1])
        self.k_face.append(indices_added[2])
        if outline:
            self.add_line(list(points) + [points[0]])
        if mirror:
            reflected_points = [reflect_over_XZ_plane(point) for point in points]
            self.add_tri(
                points=reflected_points,
                intensity=intensity,
                outline=outline,
                mirror=False
            )

    def add_quad(self,
                 points,
                 intensity=0,
                 outline=True,
                 mirror=False,
                 ):
        """
        Adds a quadrilateral face to draw. All points should be (approximately) coplanar if you want it to look right.
        :param points: an iterable with 4 items. Each item is a 3D point, represented as an iterable of length 3. Points should be given in sequential order.
        :param intensity: Intensity associated with this face
        :param outline: Do you want to outline this quad? [boolean]
        :param mirror: Should we also draw a version that's mirrored over the XZ plane? [boolean]
        :return: None

        E.g. add_face([(0, 0, 0), (1, 0, 0), (0, 1, 0)])
        """
        if not len(points) == 4:
            raise ValueError("'points' must have exactly 4 items!")
        for p in points:
            self.x_face.append(float(p[0]))
            self.y_face.append(float(p[1]))
            self.z_face.append(float(p[2]))
            self.intensity_face.append(intensity)
        indices_added = np.arange(len(self.x_face) - 4, len(self.x_face))

        self.i_face.append(indices_added[0])
        self.j_face.append(indices_added[1])
        self.k_face.append(indices_added[2])

        self.i_face.append(indices_added[0])
        self.j_face.append(indices_added[2])
        self.k_face.append(indices_added[3])

        if outline:
            self.add_line(list(points) + [points[0]])
        if mirror:
            reflected_points = [reflect_over_XZ_plane(point) for point in points]
            self.add_quad(
                points=reflected_points,
                intensity=intensity,
                outline=outline,
                mirror=False
            )

    def draw(self,
             show=True,
             title="",
             colorbar_title="",
             colorscale="viridis",
             ):
        # Draw faces
        self.fig.add_trace(
            go.Mesh3d(
                x=self.x_face,
                y=self.y_face,
                z=self.z_face,
                i=self.i_face,
                j=self.j_face,
                k=self.k_face,
                flatshading=False,
                intensity=self.intensity_face,
                colorbar=dict(title=colorbar_title),
                colorscale=colorscale,
                showscale=colorbar_title is not None
            ),
        )

        # Draw lines
        self.fig.add_trace(
            go.Scatter3d(
                x=self.x_line,
                y=self.y_line,
                z=self.z_line,
                mode='lines',
                name='',
                line=dict(color='rgb(0,0,0)', width=3),
                showlegend=False,
            )
        )

        # Draw streamlines
        self.fig.add_trace(
            go.Scatter3d(
                x=self.x_streamline,
                y=self.y_streamline,
                z=self.z_streamline,
                mode='lines',
                name='',
                line=dict(color='rgba(119,0,255,200)', width=1),
                showlegend=False,
            )
        )

        self.fig.update_layout(
            title=title,
            scene=dict(aspectmode='data'),
        )

        if show:
            self.fig.show()

        return self.fig


def spy(
        matrix,
        show=True,
):
    """
    Plots the sparsity pattern of a matrix.
    :param matrix: The matrix to plot the sparsity pattern of. [2D ndarray or CasADi array]
    :param show: Whether or not to show the sparsity plot. [boolean]
    :return: The figure to be plotted [go.Figure]
    """
    try:
        matrix = matrix.toarray()
    except:
        pass
    abs_m = np.abs(matrix)
    sparsity_pattern = abs_m >= 1e-16
    matrix[sparsity_pattern] = np.log10(abs_m[sparsity_pattern] + 1e-16)
    j_index_map, i_index_map = np.meshgrid(np.arange(matrix.shape[1]), np.arange(matrix.shape[0]))

    i_index = i_index_map[sparsity_pattern]
    j_index = j_index_map[sparsity_pattern]
    val = matrix[sparsity_pattern]
    val = np.ones_like(i_index)
    fig = go.Figure(
        data=go.Heatmap(
            y=i_index,
            x=j_index,
            z=val,
            # type='heatmap',
            colorscale='RdBu',
            showscale=False,
        ),
    )
    fig.update_layout(
        plot_bgcolor="black",
        xaxis=dict(showgrid=False, zeroline=False),
        yaxis=dict(showgrid=False, zeroline=False, autorange="reversed", scaleanchor="x", scaleratio=1),
        width=800,
        height=800 * (matrix.shape[0] / matrix.shape[1]),
    )
    if show:
        fig.show()
    return fig


def contour(
        func,  # type: callable
        x_range,  # type: tuple
        y_range,  # type: tuple
        resolution=50,  # type: int
        show=True,  # type: bool
):
    """
    Makes a contour plot of a function of 2 variables. Can also plot a list of functions.
    :param func: function of form f(x,y) to plot.
    :param x_range: Range of x values to plot, expressed as a tuple (x_min, x_max)
    :param y_range: Range of y values to plot, expressed as a tuple (y_min, y_max)
    :param resolution: Resolution in x and y to plot. [int]
    :param show: Should we show the plot?
    :return:
    """
    fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    x = np.linspace(x_range[0], x_range[1], resolution)
    y = np.linspace(y_range[0], y_range[1], resolution)



import casadi as cas


def power_human(
        duration,  # type: float
        dataset="Healthy Men"  # type: str
):
    """
    Finds the power output that a human can sustain for a given duration.
    Data was fit for durations in the range of 6 seconds to 60,000 seconds.
    Fits are modeled at: AeroSandbox/studies/HumanPower
    Data Source: Bicycling Science by D. Wilson, 2004. Figure 2.4.
        Wilson is aggregating many data sources here.
        The raw data pulls from a variety of sources:
            * NASA SP-3006, 1964
            * U.K. amateur trials and time-trials records (Whitt, F.R. 1971 "A note on the estimation of the energy expenditure of sporting cyclists." Ergonomics 14)
            * Wilsons' own analyses
    Weight estimates for tests subjects are unfortunately not given.
    :param duration: Time to sustain power output [seconds]
    :param dataset: Dataset to pull from. A string that is one of the following:
        "Healthy Men",
        "First-Class Athletes",
        "World-Class Athletes",
    :return: Sustainable power output for the specified duration [W]
    """
    if dataset == "Healthy Men":
        a = 373.153360
        b0 = -0.173127
        b1 = 0.083282
        b2 = -0.042785
    elif dataset == "First-Class Athletes":
        a = 502.332185
        b0 = -0.179030
        b1 = 0.097926
        b2 = -0.024855
    elif dataset == "World-Class Athletes":
        a = 869.963370
        b0 = -0.234291
        b1 = 0.064395
        b2 = -0.009197
    else:
        raise ValueError("Bad value of 'dataset'!")

    duration_mins = duration / 60
    log_duration_mins = cas.log10(duration_mins)

    return a * duration_mins ** (
            b0 +
            b1 * log_duration_mins +
            b2 * log_duration_mins ** 2
    ) # essentially, a cubic in log-log space

if __name__ == '__main__':
    print(
        power_human(
            duration=60,
        )
    )


from aerosandbox.geometry import *
from scipy import optimize

### Design Conditions
Re_des = 3e5  # Re to design to
Cl_start = 1.0  # Lower bound of CLs that you care about
Cl_end = 1.5  # Upper bound of CLs that you care about (Effectively, CL_max)
Cm_min = -0.08  # Worst-allowable pitching moment that you'll allow
TE_thickness = 0.0015  # Sets trailing edge thickness
enforce_continuous_LE_radius = True  # Should we force the leading edge to have continous curvature?

### Guesses for airfoil CST parameters; you usually don't need to change these
lower_guess = -0.05 * np.ones(30)
upper_guess = 0.25 * np.ones(30)
upper_guess[0] = 0.15
upper_guess[1] = 0.20

# lower_guess = [-0.21178419, -0.05500152, -0.04540216, -0.03436429, -0.03305599,
#                -0.03121454, -0.04513736, -0.05491045, -0.02861083, -0.05673649,
#                -0.06402239, -0.05963394, -0.0417384, -0.0310728, -0.04983729,
#                -0.04211283, -0.04999657, -0.0632682, -0.07226548, -0.03604782,
#                -0.06151112, -0.04030985, -0.02748867, -0.02705322, -0.04279788,
#                -0.04734922, -0.033705, -0.02380217, -0.04480772, -0.03756881]
# upper_guess = [0.17240303, 0.26668075, 0.21499604, 0.26299318, 0.22545807,
#                0.24759903, 0.31644402, 0.2964658, 0.15360716, 0.31317824,
#                0.27760982, 0.23009955, 0.24045039, 0.37542525, 0.21361931,
#                0.18678503, 0.23466624, 0.20630533, 0.16191541, 0.20453953,
#                0.14370825, 0.13428077, 0.15387739, 0.13767285, 0.15173257,
#                0.14042002, 0.11336701, 0.35640688, 0.10953915, 0.08167446]

### Packing/Unpacking functions
n_lower = len(lower_guess)
n_upper = len(upper_guess)
pack = lambda lower, upper: np.concatenate((lower, upper))
unpack = lambda pack: (pack[:n_lower], pack[n_lower:])


def make_airfoil(x):
    """
    A function that constructs an airfoil from a packed design vector.
    :param x:
    :return:
    """
    lower, upper = unpack(x)
    return Airfoil(
        name="Optimization Airfoil",
        coordinates=kulfan_coordinates(
            lower_weights=lower,
            upper_weights=upper,
            enforce_continuous_LE_radius=enforce_continuous_LE_radius,
            TE_thickness=TE_thickness,
            n_points_per_side=80
        )
    )


### Initial guess construction
x0 = pack(lower_guess, upper_guess)
initial_airfoil = make_airfoil(x0)

### Initialize plotting
fig = plt.figure(figsize=(15, 2.5))
ax = fig.add_subplot(111)
trace_initial, = ax.plot(
    initial_airfoil.coordinates[:, 0],
    initial_airfoil.coordinates[:, 1],
    ':r',
    label="Initial Airfoil"
)
trace_current, = ax.plot(
    initial_airfoil.coordinates[:, 0],
    initial_airfoil.coordinates[:, 1],
    "-b",
    label="Current Airfoil"
)
plt.axis("equal")
plt.xlabel(r"$x/c$")
plt.ylabel(r"$y/c$")
plt.title("Airfoil Optimization")
plt.legend()


def draw(
        airfoil  # type: Airfoil
):
    """
    Updates the "current airfoil" line on the plot with the given airfoil.
    :param airfoil:
    :return:
    """
    trace_current.set_xdata(airfoil.coordinates[:, 0])
    trace_current.set_ydata(airfoil.coordinates[:, 1])
    plt.draw()
    plt.pause(0.001)


### Utilities for tracking the design vector and objective throughout the optimization run
iteration = 0
xs = []
fs = []


def augmented_objective(x):
    """
    Objective function with constraints added via a multiplicative external penalty method
    :param x: Packed design vector
    :return: Value of the augmented objective
    """
    airfoil = make_airfoil(x)
    xfoil = airfoil.xfoil_cseq(
        cl_start=Cl_start,
        cl_step=0.02,
        cl_end=Cl_end,
        Re=Re_des,
        verbose=False,
        max_iter=40,
        repanel=False
    )
    if np.isnan(xfoil["Cd"]).any():
        return np.Inf

    objective = np.sqrt(np.mean(xfoil["Cd"] ** 2))  # RMS

    penalty = 0
    penalty += np.sum(np.minimum(0, (xfoil["Cm"] - Cm_min) / 0.01) ** 2)  # Cm constraint
    penalty += np.minimum(0, (airfoil.TE_angle() - 5) / 1) ** 2  # TE angle constraint
    penalty += np.minimum(0, (airfoil.local_thickness(0.90) - 0.015) / 0.005) ** 2  # Spar thickness constraint
    penalty += np.minimum(0, (airfoil.local_thickness(0.30) - 0.12) / 0.005) ** 2  # Spar thickness constraint

    xs.append(x)
    fs.append(objective)

    return objective * (1 + penalty)


def callback(x):
    global iteration
    iteration += 1
    print(
        "Iteration %i: Cd = %.6f" % (iteration, fs[-1])
    )
    if iteration % 1 == 0:
        airfoil = make_airfoil(x)
        draw(airfoil)
        ax.set_title("Airfoil Optimization: Iteration %i" % iteration)
        airfoil.write_dat("optimized_airfoil.dat")


if __name__ == '__main__':
    draw(initial_airfoil)

    initial_simplex = (
            (0.5 + 1 * np.random.random((len(x0) + 1, len(x0))))
            * x0
    )
    initial_simplex[0, :] = x0  # Include x0 in the simplex
    print("Initializing simplex (give this a few minutes)...")
    res = optimize.minimize(
        fun=augmented_objective,
        x0=pack(lower_guess, upper_guess),
        method="Nelder-Mead",
        callback=callback,
        options={
            'maxiter'        : 10 ** 6,
            'initial_simplex': initial_simplex,
            'xatol'          : 1e-8,
            'fatol'          : 1e-6,
            'adaptive'       : False,
        }
    )

    final_airfoil = make_airfoil(res.x)



from aerosandbox.geometry.common import *


class Wing(AeroSandboxObject):
    """
    Definition for a wing.
    If the wing is symmetric across the XZ plane, just define the right half and supply "symmetric = True" in the constructor.
    If the wing is not symmetric across the XZ plane, just define the wing.
    """

    def __init__(self,
                 name="Untitled Wing",  # It can help when debugging to give each wing a sensible name.
                 x_le=0,  # Will translate all of the xsecs of the wing. Useful for moving the wing around.
                 y_le=0,  # Will translate all of the xsecs of the wing. Useful for moving the wing around.
                 z_le=0,  # Will translate all of the xsecs of the wing. Useful for moving the wing around.
                 xsecs=[],  # This should be a list of WingXSec objects.
                 symmetric=False,  # Is the wing symmetric across the XZ plane?
                 chordwise_panels=8,  # The number of chordwise panels to be used in VLM and panel analyses.
                 chordwise_spacing="cosine",  # Can be 'cosine' or 'uniform'. Highly recommended to be cosine.
                 ):
        self.name = name
        self.xyz_le = cas.vertcat(x_le, y_le, z_le)
        self.xsecs = xsecs
        self.symmetric = symmetric
        self.chordwise_panels = chordwise_panels
        self.chordwise_spacing = chordwise_spacing

    def __repr__(self):
        return "Wing %s (%i xsecs, %s)" % (
            self.name,
            len(self.xsecs),
            "symmetric" if self.symmetric else "not symmetric"
        )

    def area(self,
             type="wetted"
             ):
        """
        Returns the area, with options for various ways of measuring this.
         * wetted: wetted area
         * projected: area projected onto the XY plane (top-down view)
        :param type:
        :return:
        """
        area = 0
        for i in range(len(self.xsecs) - 1):
            chord_eff = (self.xsecs[i].chord
                         + self.xsecs[i + 1].chord) / 2
            this_xyz_te = self.xsecs[i].xyz_te()
            that_xyz_te = self.xsecs[i + 1].xyz_te()
            if type == "wetted":
                span_le_eff = cas.sqrt(
                    (self.xsecs[i].xyz_le[1] - self.xsecs[i + 1].xyz_le[1]) ** 2 +
                    (self.xsecs[i].xyz_le[2] - self.xsecs[i + 1].xyz_le[2]) ** 2
                )
                span_te_eff = cas.sqrt(
                    (this_xyz_te[1] - that_xyz_te[1]) ** 2 +
                    (this_xyz_te[2] - that_xyz_te[2]) ** 2
                )
            elif type == "projected":
                span_le_eff = cas.fabs(
                    self.xsecs[i].xyz_le[1] - self.xsecs[i + 1].xyz_le[1]
                )
                span_te_eff = cas.fabs(
                    this_xyz_te[1] - that_xyz_te[1]
                )
            else:
                raise ValueError("Bad value of 'type'!")

            span_eff = (span_le_eff + span_te_eff) / 2
            area += chord_eff * span_eff
        if self.symmetric:
            area *= 2
        return area

    def span(self,
             type="wetted"
             ):
        """
        Returns the span, with options for various ways of measuring this.
         * wetted: Adds up YZ-distances of each section piece by piece
         * yz: YZ-distance between the root and tip of the wing
         * y: Y-distance between the root and tip of the wing
         * z: Z-distance between the root and tip of the wing
        If symmetric, this is doubled to obtain the full span.
        :param type: One of the above options, as a string.
        :return: span
        """
        if type == "wetted":
            span = 0
            for i in range(len(self.xsecs) - 1):
                sect1_xyz_le = self.xsecs[i].xyz_le
                sect2_xyz_le = self.xsecs[i + 1].xyz_le
                sect1_xyz_te = self.xsecs[i].xyz_te()
                sect2_xyz_te = self.xsecs[i + 1].xyz_te()

                span_le = cas.sqrt(
                    (sect1_xyz_le[1] - sect2_xyz_le[1]) ** 2 +
                    (sect1_xyz_le[2] - sect2_xyz_le[2]) ** 2
                )
                span_te = cas.sqrt(
                    (sect1_xyz_te[1] - sect2_xyz_te[1]) ** 2 +
                    (sect1_xyz_te[2] - sect2_xyz_te[2]) ** 2
                )
                span_eff = (span_le + span_te) / 2
                span += span_eff

        elif type == "yz":
            root = self.xsecs[0]  # type: WingXSec
            tip = self.xsecs[-1]  # type: WingXSec
            span = cas.sqrt(
                (root.xyz_le[1] - tip.xyz_le[1]) ** 2 +
                (root.xyz_le[2] - tip.xyz_le[2]) ** 2
            )
        elif type == "y":
            root = self.xsecs[0]  # type: WingXSec
            tip = self.xsecs[-1]  # type: WingXSec
            span = cas.fabs(
                tip.xyz_le[1] - root.xyz_le[1]
            )
        elif type == "z":
            root = self.xsecs[0]  # type: WingXSec
            tip = self.xsecs[-1]  # type: WingXSec
            span = cas.fabs(
                tip.xyz_le[2] - root.xyz_le[2]
            )
        else:
            raise ValueError("Bad value of 'type'!")
        if self.symmetric:
            span *= 2
        return span

    def aspect_ratio(self):
        # Returns the aspect ratio (b^2/S).
        # Uses the full span and the full area if symmetric.
        return self.span() ** 2 / self.area()

    def has_symmetric_control_surfaces(self):
        # Returns a boolean of whether the wing is totally symmetric (i.e.), every xsec has control_surface_type = "symmetric".
        for xsec in self.xsecs:
            if not xsec.control_surface_type == "symmetric":
                return False
        return True

    def mean_geometric_chord(self):
        """
        Returns the mean geometric chord of the wing (S/b).
        :return:
        """
        return self.area() / self.span()

    def mean_twist_angle(self):
        """
        Returns the mean twist angle (in degrees) of the wing, weighted by span.
        You can think of it as \int_{b}(twist)db, where b is span.
        WARNING: This function's output is only exact in the case where all of the cross sections have the same twist axis!
        :return: mean twist angle (in degrees)
        """
        # First, find the spans
        span = []
        for i in range(len(self.xsecs) - 1):
            sect1_xyz_le = self.xsecs[i].xyz_le
            sect2_xyz_le = self.xsecs[i + 1].xyz_le
            sect1_xyz_te = self.xsecs[i].xyz_te()
            sect2_xyz_te = self.xsecs[i + 1].xyz_te()

            span_le = cas.sqrt(
                (sect1_xyz_le[1] - sect2_xyz_le[1]) ** 2 +
                (sect1_xyz_le[2] - sect2_xyz_le[2]) ** 2
            )
            span_te = cas.sqrt(
                (sect1_xyz_te[1] - sect2_xyz_te[1]) ** 2 +
                (sect1_xyz_te[2] - sect2_xyz_te[2]) ** 2
            )
            span_eff = (span_le + span_te) / 2
            span.append(span_eff)

        # Then, find the twist-span product
        twist_span_product = 0
        for i in range(len(self.xsecs)):
            xsec = self.xsecs[i]
            if i > 0:
                twist_span_product += xsec.twist * span[i - 1] / 2
            if i < len(self.xsecs) - 1:
                twist_span_product += xsec.twist * span[i] / 2

        # Then, divide
        mean_twist = twist_span_product / cas.sum1(cas.vertcat(*span))
        return mean_twist

    def mean_sweep_angle(self):
        """
        Returns the mean quarter-chord sweep angle (in degrees) of the wing, relative to the x-axis.
        Positive sweep is backwards, negative sweep is forward.
        :return:
        """
        root_quarter_chord = 0.75 * self.xsecs[0].xyz_le + 0.25 * self.xsecs[0].xyz_te()
        tip_quarter_chord = 0.75 * self.xsecs[-1].xyz_le + 0.25 * self.xsecs[-1].xyz_te()

        vec = tip_quarter_chord - root_quarter_chord
        vec_norm = vec / cas.norm_2(vec)

        sin_sweep = vec_norm[0]  # from dot product with x_hat

        sweep_deg = cas.asin(sin_sweep) * 180 / cas.pi

        return sweep_deg

    def approximate_center_of_pressure(self):
        """
        Returns the approximate location of the center of pressure. Given as the area-weighted quarter chord of the wing.
        :return: [x, y, z] of the approximate center of pressure
        """
        areas = []
        quarter_chord_centroids = []
        for i in range(len(self.xsecs) - 1):
            # Find areas
            chord_eff = (self.xsecs[i].chord
                         + self.xsecs[i + 1].chord) / 2
            this_xyz_te = self.xsecs[i].xyz_te()
            that_xyz_te = self.xsecs[i + 1].xyz_te()
            span_le_eff = cas.sqrt(
                (self.xsecs[i].xyz_le[1] - self.xsecs[i + 1].xyz_le[1]) ** 2 +
                (self.xsecs[i].xyz_le[2] - self.xsecs[i + 1].xyz_le[2]) ** 2
            )
            span_te_eff = cas.sqrt(
                (this_xyz_te[1] - that_xyz_te[1]) ** 2 +
                (this_xyz_te[2] - that_xyz_te[2]) ** 2
            )
            span_eff = (span_le_eff + span_te_eff) / 2

            areas.append(chord_eff * span_eff)

            # Find quarter-chord centroids of each section
            quarter_chord_centroids.append(
                (
                        0.75 * self.xsecs[i].xyz_le + 0.25 * self.xsecs[i].xyz_te() +
                        0.75 * self.xsecs[i + 1].xyz_le + 0.25 * self.xsecs[i + 1].xyz_te()
                ) / 2 + self.xyz_le
            )

        areas = cas.vertcat(*areas)
        quarter_chord_centroids = cas.transpose(cas.horzcat(*quarter_chord_centroids))

        total_area = cas.sum1(areas)
        approximate_cop = cas.sum1(areas / cas.sum1(areas) * quarter_chord_centroids)

        if self.symmetric:
            approximate_cop[:, 1] = 0

        return approximate_cop


class WingXSec(AeroSandboxObject):
    """
    Definition for a wing cross section ("X-section").
    """

    def __init__(self,
                 x_le=0,  # Coordinate of the leading edge of the cross section, relative to the wing's datum.
                 y_le=0,  # Coordinate of the leading edge of the cross section, relative to the wing's datum.
                 z_le=0,  # Coordinate of the leading edge of the cross section, relative to the wing's datum.
                 chord=0,  # Chord of the wing at this cross section
                 twist=0,  # Twist always occurs about the leading edge!
                 twist_axis=cas.DM([0, 1, 0]),  # By default, always twists about the Y-axis.
                 airfoil=None,  # type: Airfoil # The airfoil to be used at this cross section.
                 control_surface_type="symmetric",
                 # Can be "symmetric" or "asymmetric". Symmetric is like flaps, asymmetric is like an aileron.
                 control_surface_hinge_point=0.75,  # The location of the control surface hinge, as a fraction of chord.
                 # Point at which the control surface is applied, as a fraction of chord.
                 control_surface_deflection=0,  # Control deflection, in degrees. Downwards-positive.
                 spanwise_panels=8,
                 # The number of spanwise panels to be used between this cross section and the next one.
                 spanwise_spacing="cosine"  # Can be 'cosine' or 'uniform'. Highly recommended to be cosine.
                 ):
        if airfoil is None:
            raise ValueError("'airfoil' argument missing! (Needs an object of Airfoil type)")

        self.x_le = x_le
        self.y_le = y_le
        self.z_le = z_le

        self.chord = chord
        self.twist = twist
        self.twist_axis = twist_axis
        self.airfoil = airfoil
        self.control_surface_type = control_surface_type
        self.control_surface_hinge_point = control_surface_hinge_point
        self.control_surface_deflection = control_surface_deflection
        self.spanwise_panels = spanwise_panels
        self.spanwise_spacing = spanwise_spacing

        self.xyz_le = cas.vertcat(x_le, y_le, z_le)

    def __repr__(self):
        return "WingXSec (airfoil = %s, chord = %f, twist = %f)" % (
            self.airfoil.name,
            self.chord,
            self.twist
        )

    def xyz_te(self):
        """
        Returns the (wing-relative) coordinates of the trailing edge of the cross section.
        """
        rot = angle_axis_rotation_matrix(self.twist * cas.pi / 180, self.twist_axis)
        xyz_te = self.xyz_le + rot @ cas.vertcat(self.chord, 0, 0)
        # xyz_te = self.xyz_le + self.chord * cas.vertcat(
        #     cas.cos(self.twist * cas.pi / 180),
        #     0,
        #     -cas.sin(self.twist * cas.pi / 180)
        # )
        return xyz_te



from aerosandbox.geometry.common import *
from aerosandbox.tools.airfoil_fitter.airfoil_fitter import AirfoilFitter
from scipy.interpolate import interp1d


class Airfoil:
    def __init__(self,
                 name=None,  # Examples: 'naca0012', 'ag10', 's1223', or anything you want.
                 coordinates=None,  # Treat this as an immutable, don't edit directly after initialization.
                 CL_function=None,  # lambda alpha, Re, mach, deflection,: (  # Lift coefficient function (alpha in deg)
                 # (alpha * np.pi / 180) * (2 * np.pi)
                 # ),  # type: callable # with exactly the arguments listed (no more, no fewer).
                 CDp_function=None,
                 # lambda alpha, Re, mach, deflection: (  # Profile drag coefficient function (alpha in deg)
                 # (1 + (alpha / 5) ** 2) * 2 * (0.074 / Re ** 0.2)
                 # ),  # type: callable # with exactly the arguments listed (no more, no fewer).
                 Cm_function=None,  # lambda alpha, Re, mach, deflection: (
                 # Moment coefficient function (about quarter-chord) (alpha in deg)
                 # 0
                 # ),  # type: callable # with exactly the arguments listed (no more, no fewer).
                 repanel=False
                 ):
        """
        Creates an Airfoil object.
        :param name: Name of the airfoil [string]
        :param coordinates: Either:
            a) None if "name" is a 4-digit NACA airfoil (e.g. "naca2412"),
            a) None if "name" is the name of an airfoil in the UIUC airfoil database (must be the name of the .dat file, e.g. "s1223"),
            b) a filepath to a .dat file (including the .dat) [string], or
            c) an array of coordinates [Nx2 ndarray].
        :param CL_function:
        :param CDp_function:
        :param Cm_function:
        :param repanel: should we repanel this airfoil upon creation?
        """
        self.name = name if name is not None else "Untitled"
        self.coordinates = None
        if coordinates is not None:
            if type(coordinates) is str:  # Assume coordinates is a filepath to a .dat file
                self.populate_coordinates_from_filepath(filepath=coordinates)
            else:  # Assume coordinates are the actual coordinates
                self.coordinates = coordinates
        else:  # There are no coordinates given
            try:  # See if it's a NACA airfoil
                self.populate_coordinates_from_naca()
            except:
                try:  # See if it's in the UIUC airfoil database
                    self.populate_coordinates_from_UIUC_database()
                except:
                    pass

        self.CL_function = CL_function
        self.CDp_function = CDp_function
        self.Cm_function = Cm_function

        if repanel:
            self.repanel(inplace=True)

    def __repr__(self):
        return "Airfoil %s (%i points)" % (
            self.name,
            self.coordinates.shape[0] if self.coordinates is not None else 0,
        )

    def populate_coordinates_from_naca(self, n_points_per_side=100):
        """
        Populates a variable called self.coordinates with the coordinates of the airfoil.
        :param n_points_per_side: Number of points per side of the airfoil (top/bottom).
        :return: None (in-place)
        """
        name = self.name.lower().strip()
        assert "naca" in name, "Not a NACA airfoil!"

        nacanumber = name.split("naca")[1]
        assert nacanumber.isdigit(), "Couldn't parse the number of the NACA airfoil!"

        assert len(nacanumber) == 4, "Can only parse 4-digit NACA airfoils at the moment!"

        # Parse
        max_camber = int(nacanumber[0]) * 0.01
        camber_loc = int(nacanumber[1]) * 0.1
        thickness = int(nacanumber[2:]) * 0.01

        # Referencing https://en.wikipedia.org/wiki/NACA_airfoil#Equation_for_a_cambered_4-digit_NACA_airfoil
        # from here on out

        # Make uncambered coordinates
        x_t = cosspace(0, 1, n_points_per_side)  # Generate some cosine-spaced points
        y_t = 5 * thickness * (
                + 0.2969 * x_t ** 0.5
                - 0.1260 * x_t
                - 0.3516 * x_t ** 2
                + 0.2843 * x_t ** 3
                - 0.1015 * x_t ** 4  # 0.1015 is original, #0.1036 for sharp TE
        )

        if camber_loc == 0:
            camber_loc = 0.5  # prevents divide by zero errors for things like naca0012's.

        # Get camber

        y_c = cas.if_else(
            x_t <= camber_loc,
            max_camber / camber_loc ** 2 * (2 * camber_loc * x_t - x_t ** 2),
            max_camber / (1 - camber_loc) ** 2 * ((1 - 2 * camber_loc) + 2 * camber_loc * x_t - x_t ** 2)
        )

        # Get camber slope
        dycdx = cas.if_else(
            x_t <= camber_loc,
            2 * max_camber / camber_loc ** 2 * (camber_loc - x_t),
            2 * max_camber / (1 - camber_loc) ** 2 * (camber_loc - x_t)
        )
        theta = cas.atan(dycdx)

        # Combine everything
        x_U = x_t - y_t * cas.sin(theta)
        x_L = x_t + y_t * cas.sin(theta)
        y_U = y_c + y_t * cas.cos(theta)
        y_L = y_c - y_t * cas.cos(theta)

        # Flip upper surface so it's back to front
        x_U, y_U = x_U[::-1, :], y_U[::-1, :]

        # Trim 1 point from lower surface so there's no overlap
        x_L, y_L = x_L[1:], y_L[1:]

        x = cas.vertcat(x_U, x_L)
        y = cas.vertcat(y_U, y_L)

        self.coordinates = np.array(cas.horzcat(x, y))

    def populate_coordinates_from_UIUC_database(self):
        """
        Populates a variable called self.coordinates with the coordinates of the airfoil.
        :return: None (in-place)
        """

        name = self.name.lower().strip()

        import importlib.resources
        from aerosandbox import airfoil_database

        try:
            with importlib.resources.open_text(airfoil_database, name) as f:
                raw_text = f.readlines()
        except:
            with importlib.resources.open_text(airfoil_database, name + '.dat') as f:
                raw_text = f.readlines()

        trimmed_text = []
        for line in raw_text:
            try:
                line_np = np.fromstring(line, sep=" ")
                if line_np.shape[0] == 2:
                    trimmed_text.append(line_np)
            except:
                pass

        coordinates = np.hstack(trimmed_text).reshape((-1, 2))
        self.coordinates = coordinates

    def populate_coordinates_from_filepath(self, filepath):
        """
        Populates a variable called self.coordinates with the coordinates of the airfoil.
        :param filepath: A DAT file to pull the airfoil coordinates from. (includes the ".dat") [string]
        :return: None (in-place)
        """
        try:
            with open(filepath, "r") as f:
                raw_text = f.readlines()
        except:
            with open(filepath + ".dat", "r") as f:
                raw_text = f.readlines()

        trimmed_text = []
        for line in raw_text:
            try:
                line_np = np.fromstring(line, sep=" ")
                if line_np.shape[0] == 2:
                    trimmed_text.append(line_np)
            except:
                pass

        coordinates = np.hstack(trimmed_text).reshape((-1, 2))
        self.coordinates = coordinates

    def populate_sectional_functions_from_xfoil_fits(self,
                                                     parallel=True,
                                                     verbose=True,
                                                     ):  # TODO write docstring
        if not self.has_xfoil_data(raise_exception_if_absent=False):
            self.get_xfoil_data(
                parallel=parallel,
                verbose=verbose,
            )

        self.AirfoilFitter = AirfoilFitter(
            airfoil=self,
            parallel=parallel,
            verbose=verbose,
        )
        self.AirfoilFitter.fit_xfoil_data_Cl(plot_fit=False)
        self.AirfoilFitter.fit_xfoil_data_Cd(plot_fit=False)

        def CL_function(
                alpha, Re, mach=0, deflection=0,
                fitter=self.AirfoilFitter
        ):
            return fitter.Cl_function(
                alpha=alpha,
                Re=Re,
            )

        def CDp_function(
                alpha, Re, mach=0, deflection=0,
                fitter=self.AirfoilFitter
        ):
            return fitter.Cd_function(
                alpha=alpha,
                Re=Re,
            )

        def Cm_function(
                alpha, Re, mach=0, deflection=0,
                fitter=self.AirfoilFitter
        ):
            return alpha * 0

        self.CL_function = CL_function
        self.CDp_function = CDp_function
        self.Cm_function = Cm_function

    def has_sectional_functions(self, raise_exception_if_absent=True):
        """
        Runs a quick check to see if this airfoil has sectional functions.
        :param raise_exception_if_absent: Boolean flag to raise an Exception if sectional functions are not found.
        :return: Boolean of whether or not sectional functions is present.
        """
        data_present = (
                hasattr(self, 'CL_function') and callable(self.CL_function) and
                hasattr(self, 'CDp_function') and callable(self.CDp_function) and
                hasattr(self, 'Cm_function') and callable(self.Cm_function)
        )
        if not data_present and raise_exception_if_absent:
            raise Exception(
                """This Airfoil %s does not yet have sectional functions,
                so you can't run the function you've called.
                To get sectional functions, first call:
                    Airfoil.populate_sectional_functions_from_xfoil_fits()
                which will perform an in-place update that
                provides the data.""" % self.name
            )
        return data_present

    def local_camber(self, x_over_c=np.linspace(0, 1, 101)):
        """
        Returns the local camber of the airfoil.
        :param x_over_c: The x/c locations to calculate the camber at [1D array, more generally, an iterable of floats]
        :return: Local camber of the airfoil (y/c) [1D array].
        """
        # TODO casadify?
        upper = self.upper_coordinates()[::-1]
        lower = self.lower_coordinates()

        upper_interpolated = np.interp(
            x_over_c,
            upper[:, 0],
            upper[:, 1],
        )
        lower_interpolated = np.interp(
            x_over_c,
            lower[:, 0],
            lower[:, 1],
        )

        return (upper_interpolated + lower_interpolated) / 2

    def local_thickness(self, x_over_c=np.linspace(0, 1, 101)):
        """
        Returns the local thickness of the airfoil.
        :param x_over_c: The x/c locations to calculate the thickness at [1D array, more generally, an iterable of floats]
        :return: Local thickness of the airfoil (y/c) [1D array].
        """
        # TODO casadify?
        upper = self.upper_coordinates()[::-1]
        lower = self.lower_coordinates()

        upper_interpolated = np.interp(
            x_over_c,
            upper[:, 0],
            upper[:, 1],
        )
        lower_interpolated = np.interp(
            x_over_c,
            lower[:, 0],
            lower[:, 1],
        )

        return upper_interpolated - lower_interpolated

    def draw(self, draw_mcl=True, backend="plotly"):
        """
        Draw the airfoil object.
        :param draw_mcl: Should we draw the mean camber line (MCL)? [boolean]
        :param backend: Which backend should we use? "plotly" or "matplotlib" [boolean]
        :return: None
        """
        x = np.array(self.coordinates[:, 0]).reshape(-1)
        y = np.array(self.coordinates[:, 1]).reshape(-1)
        if backend == "plotly":
            fig = go.Figure()
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=y,
                    mode="lines+markers",
                    name="Airfoil"
                ),
            )
            if draw_mcl:
                x_mcl = np.linspace(np.min(x), np.max(x), len(x))
                y_mcl = self.local_camber(x_mcl)
                fig.add_trace(
                    go.Scatter(
                        x=x_mcl,
                        y=y_mcl,
                        mode="lines+markers",
                        name="Mean Camber Line (MCL)"
                    )
                )
            fig.update_layout(
                xaxis_title="x/c",
                yaxis_title="y/c",
                yaxis=dict(scaleanchor="x", scaleratio=1),
                title="%s Airfoil" % self.name
            )
            fig.show()
        elif backend == "matplotlib":
            fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
            plt.plot(x, y, ".-", zorder=11, color='#280887')
            if draw_mcl:
                x_mcl = np.linspace(np.min(x), np.max(x), len(x))
                y_mcl = self.local_camber(x_mcl)
                plt.plot(x_mcl, y_mcl, "-", zorder=4, color='#28088744')
            plt.axis("equal")
            plt.xlabel(r"$x/c$")
            plt.ylabel(r"$y/c$")
            plt.title("%s Airfoil" % self.name)
            plt.tight_layout()
            plt.show()

    def LE_index(self):
        # Returns the index of the leading-edge point.
        return np.argmin(self.coordinates[:, 0])

    def lower_coordinates(self):
        # Returns a matrix (N by 2) of [x, y] coordinates that describe the lower surface of the airfoil.
        # Order is from leading edge to trailing edge.
        # Includes the leading edge point; be careful about duplicates if using this method in conjunction with self.upper_coordinates().
        return self.coordinates[self.LE_index():, :]

    def upper_coordinates(self):
        # Returns a matrix (N by 2) of [x, y] coordinates that describe the upper surface of the airfoil.
        # Order is from trailing edge to leading edge.
        # Includes the leading edge point; be careful about duplicates if using this method in conjunction with self.lower_coordinates().
        return self.coordinates[:self.LE_index() + 1, :]

    def TE_thickness(self):
        # Returns the thickness of the trailing edge of the airfoil, in nondimensional (chord-normalized) units.
        return self.local_thickness(x_over_c=1)

    def TE_angle(self):
        # Returns the trailing edge angle of the airfoil, in degrees
        upper_TE_vec = self.coordinates[0, :] - self.coordinates[1, :]
        lower_TE_vec = self.coordinates[-1, :] - self.coordinates[-2, :]

        return 180 / np.pi * (np.arctan2(
            upper_TE_vec[0] * lower_TE_vec[1] - upper_TE_vec[1] * lower_TE_vec[0],
            upper_TE_vec[0] * lower_TE_vec[0] + upper_TE_vec[1] * upper_TE_vec[1]
        ))

    def area(self):
        # Returns the area of the airfoil, in nondimensional (normalized to chord^2) units.
        x = self.coordinates[:, 0]
        y = self.coordinates[:, 1]
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        return A

    def centroid(self):
        # Returns the centroid of the airfoil, in nondimensional (chord-normalized) units.
        x = self.coordinates[:, 0]
        y = self.coordinates[:, 1]
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        x_c = 1 / (6 * A) * np.sum(a * (x + x_n))
        y_c = 1 / (6 * A) * np.sum(a * (y + y_n))
        centroid = np.array([x_c, y_c])

        return centroid

    def Ixx(self):
        # Returns the nondimensionalized Ixx moment of inertia, taken about the centroid.
        x = self.coordinates[:, 0]
        y = self.coordinates[:, 1]
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        x_c = 1 / (6 * A) * cas.sum1(a * (x + x_n))
        y_c = 1 / (6 * A) * cas.sum1(a * (y + y_n))
        centroid = np.array([x_c, y_c])

        Ixx = 1 / 12 * np.sum(a * (y ** 2 + y * y_n + y_n ** 2))

        Iuu = Ixx - A * centroid[1] ** 2

        return Iuu

    def Iyy(self):
        # Returns the nondimensionalized Iyy moment of inertia, taken about the centroid.
        x = self.coordinates[:, 0]
        y = self.coordinates[:, 1]
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        x_c = 1 / (6 * A) * np.sum(a * (x + x_n))
        y_c = 1 / (6 * A) * np.sum(a * (y + y_n))
        centroid = np.array([x_c, y_c])

        Iyy = 1 / 12 * np.sum(a * (x ** 2 + x * x_n + x_n ** 2))

        Ivv = Iyy - A * centroid[0] ** 2

        return Ivv

    def Ixy(self):
        # Returns the nondimensionalized product of inertia, taken about the centroid.
        x = self.coordinates[:, 0]
        y = self.coordinates[:, 1]
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        x_c = 1 / (6 * A) * np.sum(a * (x + x_n))
        y_c = 1 / (6 * A) * np.sum(a * (y + y_n))
        centroid = np.array([x_c, y_c])

        Ixy = 1 / 24 * np.sum(a * (x * y_n + 2 * x * y + 2 * x_n * y_n + x_n * y))

        Iuv = Ixy - A * centroid[0] * centroid[1]

        return Iuv

    def J(self):
        # Returns the nondimensionalized polar moment of inertia, taken about the centroid.
        x = self.coordinates[:, 0]
        y = self.coordinates[:, 1]
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        x_c = 1 / (6 * A) * np.sum(a * (x + x_n))
        y_c = 1 / (6 * A) * np.sum(a * (y + y_n))
        centroid = np.array([x_c, y_c])

        Ixx = 1 / 12 * np.sum(a * (y ** 2 + y * y_n + y_n ** 2))

        Iyy = 1 / 12 * np.sum(a * (x ** 2 + x * x_n + x_n ** 2))

        J = Ixx + Iyy

        return J

    def repanel(self,
                n_points_per_side=80,
                inplace=False,
                ):
        """
        Returns a repaneled version of the airfoil with cosine-spaced coordinates on the upper and lower surfaces.
        :param n_points_per_side: Number of points per side (upper and lower) of the airfoil [int]
            Notes: The number of points defining the final airfoil will be n_points_per_side*2-1,
            since one point (the leading edge point) is shared by both the upper and lower surfaces.
        :param inplace: Whether to perform this as an in-place operation or return the new airfoil as a newly instantiated object [boolean]
        :return: If inplace is True, None. If inplace is False, the new airfoil [Airfoil].
        """

        upper_original_coors = self.upper_coordinates()  # Note: includes leading edge point, be careful about duplicates
        lower_original_coors = self.lower_coordinates()  # Note: includes leading edge point, be careful about duplicates

        # Find distances between coordinates, assuming linear interpolation
        upper_distances_between_points = (
                                                 (upper_original_coors[:-1, 0] - upper_original_coors[1:, 0]) ** 2 +
                                                 (upper_original_coors[:-1, 1] - upper_original_coors[1:, 1]) ** 2
                                         ) ** 0.5
        lower_distances_between_points = (
                                                 (lower_original_coors[:-1, 0] - lower_original_coors[1:, 0]) ** 2 +
                                                 (lower_original_coors[:-1, 1] - lower_original_coors[1:, 1]) ** 2
                                         ) ** 0.5
        upper_distances_from_TE = np.hstack((0, np.cumsum(upper_distances_between_points)))
        lower_distances_from_LE = np.hstack((0, np.cumsum(lower_distances_between_points)))
        upper_distances_from_TE_normalized = upper_distances_from_TE / upper_distances_from_TE[-1]
        lower_distances_from_LE_normalized = lower_distances_from_LE / lower_distances_from_LE[-1]

        distances_from_TE_normalized = np.hstack((
            upper_distances_from_TE_normalized,
            1 + lower_distances_from_LE_normalized[1:]
        ))

        # Generate a cosine-spaced list of points from 0 to 1
        cosspaced_points = np_cosspace(0, 1, n_points_per_side)
        s = np.hstack((
            cosspaced_points,
            1 + cosspaced_points[1:],
        ))

        x_coors = interp1d(
            distances_from_TE_normalized,
            self.coordinates[:, 0],
            kind="cubic",
        )(s)
        y_coors = interp1d(
            distances_from_TE_normalized,
            self.coordinates[:, 1],
            kind="cubic",
        )(s)

        coordinates = np.vstack((x_coors, y_coors)).T

        # Finalize
        airfoil = self if inplace else copy.deepcopy(self)
        if not "Repaneled" in airfoil.name:
            airfoil.name += " (Repaneled)"
        airfoil.coordinates = coordinates
        return airfoil

    def add_control_surface(
            self,
            deflection=0.,
            hinge_point_x=0.75,
            inplace=False,
    ):
        """
        Returns a version of the airfoil with a control surface added at a given point. Implicitly repanels the airfoil as part of this operation.
        :param deflection: deflection angle [degrees]. Downwards-positive.
        :param hinge_point_x: location of the hinge, as a fraction of chord [float].
        :param inplace: Whether to perform this as an in-place operation or return the new airfoil as a newly instantiated object [boolean]
        :return: If inplace is True, None. If inplace is False, the new airfoil [Airfoil].
        """

        # Make the rotation matrix for the given angle.
        sintheta = np.sin(-cas.pi / 180 * deflection)
        costheta = np.cos(-cas.pi / 180 * deflection)
        rotation_matrix = np.array([
            [costheta, sintheta],
            [-sintheta, costheta]
        ])

        # Find the hinge point
        hinge_point_y = self.local_camber(hinge_point_x)
        hinge_point = np.hstack((hinge_point_x, hinge_point_y))

        # Find the new coordinates
        c = np.copy(self.coordinates)
        c[c[:, 0] > hinge_point_x] = (rotation_matrix.T @ (c[c[:, 0] > hinge_point_x] - hinge_point).T).T + hinge_point
        coordinates = c

        # Finalize
        airfoil = self if inplace else copy.deepcopy(self)
        if not "Flapped" in airfoil.name:
            airfoil.name += " (Flapped)"
        airfoil.coordinates = coordinates
        return airfoil

    def write_dat(self,
                  filepath  # type: str
                  ):
        """
        Writes a .dat file corresponding to this airfoil to a filepath.
        :param filepath: filepath (including the filename and .dat extension) [string]
        :return: None
        """
        with open(filepath, "w+") as f:
            f.writelines(
                [self.name + "\n"] +
                ["     %.12f    %.12f\n" % tuple(coordinate) for coordinate in self.coordinates]
            )

    def xfoil_a(self,
                alpha,
                Re=0,
                M=0,
                n_crit=9,
                xtr_bot=1,
                xtr_top=1,
                reset_bls=False,
                repanel=False,
                max_iter=20,
                verbose=False,
                ):
        """
        Interface to XFoil, provided through the open-source xfoil Python library by DARcorporation.
        Point analysis at a given alpha.
        :param alpha: angle of attack [deg]
        :param Re: Reynolds number
        :param M: Mach number
        :param n_crit: Critical Tollmien-Schlichting wave amplification factor
        :param xtr_bot: Bottom trip location [x/c]
        :param xtr_top: Top trip location [x/c]
        :param reset_bls: Reset boundary layer parameters upon initialization?
        :param repanel: Repanel airfoil within XFoil?
        :param max_iter: Maximum number of global Newton iterations
        :param verbose: Choose whether you want to suppress output from xfoil [boolean]
        :return: A dict of {alpha, Cl, Cd, Cm, Cp_min}
        """
        try:
            xf = XFoil()
        except NameError:
            raise NameError(
                "It appears that the XFoil-Python interface is not installed, so unfortunately you can't use this function!\n"
                "To install it, run \"pip install xfoil\" in your terminal, or manually install it from: https://github.com/DARcorporation/xfoil-python .\n"
                "Note: users on UNIX systems have reported errors with installing this (Windows seems fine).")

        def run():
            xf.airfoil = xfoil_model.Airfoil(
                x=np.array(self.coordinates[:, 0]).reshape(-1),
                y=np.array(self.coordinates[:, 1]).reshape(-1),
            )
            xf.Re = Re
            xf.M = M
            xf.n_crit = n_crit
            xf.xtr = (xtr_top, xtr_bot)
            if reset_bls:
                xf.reset_bls()
            if repanel:
                xf.repanel()
            xf.max_iter = max_iter
            return xf.a(alpha)

        if verbose:
            cl, cd, cm, Cp_min = run()
        else:
            with stdout_redirected():
                cl, cd, cm, Cp_min = run()
        a = alpha

        return {
            "alpha" : a,
            "Cl"    : cl,
            "Cd"    : cd,
            "Cm"    : cm,
            "Cp_min": Cp_min
        }

    def xfoil_cl(self,
                 cl,
                 Re=0,
                 M=0,
                 n_crit=9,
                 xtr_bot=1,
                 xtr_top=1,
                 reset_bls=False,
                 repanel=False,
                 max_iter=20,
                 verbose=False,
                 ):
        """
        Interface to XFoil, provided through the open-source xfoil Python library by DARcorporation.
        Point analysis at a given lift coefficient.
        :param cl: Lift coefficient
        :param Re: Reynolds number
        :param M: Mach number
        :param n_crit: Critical Tollmien-Schlichting wave amplification factor
        :param xtr_bot: Bottom trip location [x/c]
        :param xtr_top: Top trip location [x/c]
        :param reset_bls: Reset boundary layer parameters upon initialization?
        :param repanel: Repanel airfoil within XFoil?
        :param max_iter: Maximum number of global Newton iterations
        :param verbose: Choose whether you want to suppress output from xfoil [boolean]
        :return: A dict of {alpha, Cl, Cd, Cm, Cp_min}
        """
        try:
            xf = XFoil()
        except NameError:
            raise NameError(
                "It appears that the XFoil-Python interface is not installed, so unfortunately you can't use this function!\n"
                "To install it, run \"pip install xfoil\" in your terminal, or manually install it from: https://github.com/DARcorporation/xfoil-python .\n"
                "Note: users on UNIX systems have reported errors with installing this (Windows seems fine).")

        def run():
            xf.airfoil = xfoil_model.Airfoil(
                x=np.array(self.coordinates[:, 0]).reshape(-1),
                y=np.array(self.coordinates[:, 1]).reshape(-1),
            )
            xf.Re = Re
            xf.M = M
            xf.n_crit = n_crit
            xf.xtr = (xtr_top, xtr_bot)
            if reset_bls:
                xf.reset_bls()
            if repanel:
                xf.repanel()
            xf.max_iter = max_iter
            return xf.cl(cl)

        if verbose:
            a, cd, cm, Cp_min = run()
        else:
            with stdout_redirected():
                a, cd, cm, Cp_min = run()

        cl = cl

        return {
            "alpha" : a,
            "Cl"    : cl,
            "Cd"    : cd,
            "Cm"    : cm,
            "Cp_min": Cp_min
        }

    def xfoil_aseq(self,
                   a_start,
                   a_end,
                   a_step,
                   Re=0,
                   M=0,
                   n_crit=9,
                   xtr_bot=1,
                   xtr_top=1,
                   reset_bls=False,
                   repanel=False,
                   max_iter=20,
                   verbose=False,
                   ):
        """
        Interface to XFoil, provided through the open-source xfoil Python library by DARcorporation.
        Alpha sweep analysis.
        :param a_start: First angle of attack [deg]
        :param a_end: Last angle of attack [deg]
        :param a_step: Amount to increment angle of attack by [deg]
        :param Re: Reynolds number
        :param M: Mach number
        :param n_crit: Critical Tollmien-Schlichting wave amplification factor
        :param xtr_bot: Bottom trip location [x/c]
        :param xtr_top: Top trip location [x/c]
        :param reset_bls: Reset boundary layer parameters upon initialization?
        :param repanel: Repanel airfoil within XFoil?
        :param max_iter: Maximum number of global Newton iterations
        :param verbose: Choose whether you want to suppress output from xfoil [boolean]
        :return: A dict of {alpha, Cl, Cd, Cm, Cp_min}
        """
        try:
            xf = XFoil()
        except NameError:
            raise NameError(
                "It appears that the XFoil-Python interface is not installed, so unfortunately you can't use this function!\n"
                "To install it, run \"pip install xfoil\" in your terminal, or manually install it from: https://github.com/DARcorporation/xfoil-python .\n"
                "Note: users on UNIX systems have reported errors with installing this (Windows seems fine).")

        def run():
            xf.airfoil = xfoil_model.Airfoil(
                x=np.array(self.coordinates[:, 0]).reshape(-1),
                y=np.array(self.coordinates[:, 1]).reshape(-1),
            )
            xf.Re = Re
            xf.M = M
            xf.n_crit = n_crit
            xf.xtr = (xtr_top, xtr_bot)
            if reset_bls:
                xf.reset_bls()
            if repanel:
                xf.repanel()
            xf.max_iter = max_iter
            return xf.aseq(a_start, a_end, a_step)

        if verbose:
            a, cl, cd, cm, Cp_min = run()
        else:
            with stdout_redirected():
                a, cl, cd, cm, Cp_min = run()

        return {
            "alpha" : a,
            "Cl"    : cl,
            "Cd"    : cd,
            "Cm"    : cm,
            "Cp_min": Cp_min
        }

    def xfoil_cseq(self,
                   cl_start,
                   cl_end,
                   cl_step,
                   Re=0,
                   M=0,
                   n_crit=9,
                   xtr_bot=1,
                   xtr_top=1,
                   reset_bls=False,
                   repanel=False,
                   max_iter=20,
                   verbose=False,
                   ):
        """
        Interface to XFoil, provided through the open-source xfoil Python library by DARcorporation.
        Lift coefficient sweep analysis.
        :param cl_start: First lift coefficient [unitless]
        :param cl_end: Last lift coefficient [unitless]
        :param cl_step: Amount to increment lift coefficient by [unitless]
        :param Re: Reynolds number
        :param M: Mach number
        :param n_crit: Critical Tollmien-Schlichting wave amplification factor
        :param xtr_bot: Bottom trip location [x/c]
        :param xtr_top: Top trip location [x/c]
        :param reset_bls: Reset boundary layer parameters upon initialization?
        :param repanel: Repanel airfoil within XFoil?
        :param max_iter: Maximum number of global Newton iterations
        :param verbose: Choose whether you want to suppress output from xfoil [boolean]
        :return: A dict of {alpha, Cl, Cd, Cm, Cp_min}
        """
        try:
            xf = XFoil()
        except NameError:
            raise NameError(
                "It appears that the XFoil-Python interface is not installed, so unfortunately you can't use this function!\n"
                "To install it, run \"pip install xfoil\" in your terminal, or manually install it from: https://github.com/DARcorporation/xfoil-python .\n"
                "Note: users on UNIX systems have reported errors with installing this (Windows seems fine).")

        def run():
            xf.airfoil = xfoil_model.Airfoil(
                x=np.array(self.coordinates[:, 0]).reshape(-1),
                y=np.array(self.coordinates[:, 1]).reshape(-1),
            )
            xf.Re = Re
            xf.M = M
            xf.n_crit = n_crit
            xf.xtr = (xtr_top, xtr_bot)
            if reset_bls:
                xf.reset_bls()
            if repanel:
                xf.repanel()
            xf.max_iter = max_iter
            return xf.cseq(cl_start, cl_end, cl_step)

        if verbose:
            a, cl, cd, cm, Cp_min = run()
        else:
            with stdout_redirected():
                a, cl, cd, cm, Cp_min = run()

        return {
            "alpha" : a,
            "Cl"    : cl,
            "Cd"    : cd,
            "Cm"    : cm,
            "Cp_min": Cp_min
        }

    def get_xfoil_data(self,
                       a_start=-6,  # type: float
                       a_end=12,  # type: float
                       a_step=0.5,  # type: float
                       a_init=0,  # type: float
                       Re_start=1e4,  # type: float
                       Re_end=1e7,  # type: float
                       n_Res=30,  # type: int
                       mach=0,  # type: float
                       max_iter=20,  # type: int
                       repanel=False,  # type: bool
                       parallel=True,  # type: bool
                       verbose=True,  # type: bool
                       ):
        """ # TODO finish docstring
        Calculates aerodynamic performance data for a particular airfoil with XFoil.
        Does a 2D grid sweep of the alpha-Reynolds space at a particular Mach number.
        Populates two new instance variables:
            * self.xfoil_data_1D: A dict of XFoil data at all calculated operating points (1D arrays, NaNs removed)
            * self.xfoil_data_2D: A dict of XFoil data at all calculated operating points (2D arrays, NaNs present)
        :param a_start: Lower bound of angle of attack [deg]
        :param a_end: Upper bound of angle of attack [deg]
        :param a_step: Angle of attack increment size [deg]
        :param a_init: Angle of attack to initialize runs at. Should solve easily (0 recommended) [deg]
        :param Re_start: Reynolds number to begin sweep at. [unitless]
        :param Re_end: Reynolds number to end sweep at. [unitless]
        :param n_Res: Number of Reynolds numbers to sweep. Points are log-spaced.
        :param mach: Mach number to sweep at.
        :param max_iter: Maximum number of XFoil iterations per op-point.
        :param repanel: Should we interally repanel the airfoil within XFoil before running? [boolean]
            Consider disabling this if you try to do optimization based on this data (for smoothness reasons).
            Otherwise, it's generally a good idea to leave this on.
        :param parallel: Should we run in parallel? Generally results in significant speedup, but might not run
            correctly on some machines. Disable this if it's a problem. [boolean]
        :param verbose: Should we do verbose output? [boolean]
        :return: self (in-place operation that creates self.xfoil_data_1D and self.xfoil_data_2D)
        """
        assert a_init > a_start
        assert a_init < a_end
        assert Re_start < Re_end
        assert n_Res >= 1
        assert mach >= 0

        Res = np.logspace(np.log10(Re_start), np.log10(Re_end), n_Res)

        def get_xfoil_data_at_Re(Re):

            import numpy as np  # needs to be imported here to support parallelization

            run_data_upper = self.xfoil_aseq(
                a_start=a_init + a_step,
                a_end=a_end,
                a_step=a_step,
                Re=Re,
                repanel=repanel,
                max_iter=max_iter,
                M=mach,
                reset_bls=True,
            )
            run_data_lower = self.xfoil_aseq(
                a_start=a_init,
                a_end=a_start,
                a_step=-a_step,
                Re=Re,
                repanel=repanel,
                max_iter=max_iter,
                M=mach,
                reset_bls=True,
            )
            run_data = {
                k: np.hstack((
                    run_data_lower[k][::-1],
                    run_data_upper[k]
                )) for k in run_data_upper.keys()
            }
            return run_data

        if verbose:
            print("Running XFoil sweeps on Airfoil %s..." % self.name)
            import time
            start_time = time.time()

        if not parallel:
            runs_data = [get_xfoil_data_at_Re(Re) for Re in Res]
        else:
            import multiprocess as mp
            pool = mp.Pool(mp.cpu_count())
            runs_data = pool.map(get_xfoil_data_at_Re, Res)
            pool.close()

        if verbose:
            run_time = time.time() - start_time
            print("XFoil Runtime: %.3f sec" % run_time)

        xfoil_data_2D = {}
        for k in runs_data[0].keys():
            xfoil_data_2D[k] = np.vstack([
                d[k]
                for d in runs_data
            ])
        xfoil_data_2D["Re"] = np.tile(Res, (
            xfoil_data_2D["alpha"].shape[1],
            1
        )).T
        np.place(
            arr=xfoil_data_2D["Re"],
            mask=np.isnan(xfoil_data_2D["alpha"]),
            vals=np.NaN
        )
        xfoil_data_2D["alpha_indices"] = np.arange(a_start, a_end + a_step / 2, a_step)
        xfoil_data_2D["Re_indices"] = Res

        self.xfoil_data_2D = xfoil_data_2D

        # 1-dimensionalize it and remove NaNs
        xfoil_data_1D = {
            k: remove_nans(xfoil_data_2D[k].reshape(-1))
            for k in xfoil_data_2D.keys()
        }
        self.xfoil_data_1D = xfoil_data_1D

        return self

    def has_xfoil_data(self, raise_exception_if_absent=True):
        """
        Runs a quick check to see if this airfoil has XFoil data.
        :param raise_exception_if_absent: Boolean flag to raise an Exception if XFoil data is not found.
        :return: Boolean of whether or not XFoil data is present.
        """
        data_present = (
                hasattr(self, 'xfoil_data_1D') and
                hasattr(self, 'xfoil_data_2D')
        )
        if not data_present and raise_exception_if_absent:
            raise Exception(
                """This Airfoil %s does not yet have XFoil data,
                so you can't run the function you've called.
                To get XFoil data, first call:
                    Airfoil.get_xfoil_data()
                which will perform an in-place update that
                provides the data.""" % self.name
            )
        return data_present

    def plot_xfoil_data_contours(self):  # TODO add docstring
        self.has_xfoil_data()  # Ensure data is present.
        from matplotlib import colors

        d = self.xfoil_data_1D  # data

        fig = plt.figure(figsize=(10, 8), dpi=200)

        ax = fig.add_subplot(311)
        coords = self.coordinates
        plt.plot(coords[:, 0], coords[:, 1], '.-', color='#280887')
        plt.xlabel(r"$x/c$")
        plt.ylabel(r"$y/c$")
        plt.title(r"XFoil Data for %s Airfoil" % self.name)
        plt.axis("equal")

        with plt.style.context("default"):
            ax = fig.add_subplot(323)
            x = d["Re"]
            y = d["alpha"]
            z = d["Cl"]
            levels = np.linspace(-0.5, 1.5, 21)
            norm = None
            CF = ax.tricontourf(x, y, z, levels=levels, norm=norm, cmap="plasma", extend="both")
            C = ax.tricontour(x, y, z, levels=levels, norm=norm, colors='k', extend="both", linewidths=0.5)
            cbar = plt.colorbar(CF, format='%.2f')
            cbar.set_label(r"$C_l$")
            plt.grid(False)
            plt.xlabel(r"$Re$")
            plt.ylabel(r"$\alpha$")
            plt.title(r"$C_l$ from $Re$, $\alpha$")
            ax.set_xscale('log')

            ax = fig.add_subplot(324)
            x = d["Re"]
            y = d["alpha"]
            z = d["Cd"]
            levels = np.logspace(-2.5, -1, 21)
            norm = colors.PowerNorm(gamma=1 / 2, vmin=np.min(levels), vmax=np.max(levels))
            CF = ax.tricontourf(x, y, z, levels=levels, norm=norm, cmap="plasma", extend="both")
            C = ax.tricontour(x, y, z, levels=levels, norm=norm, colors='k', extend="both", linewidths=0.5)
            cbar = plt.colorbar(CF, format='%.3f')
            cbar.set_label(r"$C_d$")
            plt.grid(False)
            plt.xlabel(r"$Re$")
            plt.ylabel(r"$\alpha$")
            plt.title(r"$C_d$ from $Re$, $\alpha$")
            ax.set_xscale('log')

            ax = fig.add_subplot(325)
            x = d["Re"]
            y = d["alpha"]
            z = d["Cl"] / d["Cd"]
            x = x[d["alpha"] >= 0]
            y = y[d["alpha"] >= 0]
            z = z[d["alpha"] >= 0]
            levels = np.logspace(1, np.log10(150), 21)
            norm = colors.PowerNorm(gamma=1 / 2, vmin=np.min(levels), vmax=np.max(levels))
            CF = ax.tricontourf(x, y, z, levels=levels, norm=norm, cmap="plasma", extend="both")
            C = ax.tricontour(x, y, z, levels=levels, norm=norm, colors='k', extend="both", linewidths=0.5)
            cbar = plt.colorbar(CF, format='%.1f')
            cbar.set_label(r"$L/D$")
            plt.grid(False)
            plt.xlabel(r"$Re$")
            plt.ylabel(r"$\alpha$")
            plt.title(r"$L/D$ from $Re$, $\alpha$")
            ax.set_xscale('log')

            ax = fig.add_subplot(326)
            x = d["Re"]
            y = d["alpha"]
            z = d["Cm"]
            levels = np.linspace(-0.15, 0, 21)  # np.logspace(1, np.log10(150), 21)
            norm = None  # colors.PowerNorm(gamma=1 / 2, vmin=np.min(levels), vmax=np.max(levels))
            CF = ax.tricontourf(x, y, z, levels=levels, norm=norm, cmap="plasma", extend="both")
            C = ax.tricontour(x, y, z, levels=levels, norm=norm, colors='k', extend="both", linewidths=0.5)
            cbar = plt.colorbar(CF, format='%.2f')
            cbar.set_label(r"$C_m$")
            plt.grid(False)
            plt.xlabel(r"$Re$")
            plt.ylabel(r"$\alpha$")
            plt.title(r"$C_m$ from $Re$, $\alpha$")
            ax.set_xscale('log')

        plt.tight_layout()
        plt.show()

        return self

    def plot_xfoil_data_all_polars(self,
                                   n_lines_max=20,
                                   Cd_plot_max=0.04,
                                   ):
        """
        Plots the existing XFoil data found by running self.get_xfoil_data().
        :param n_lines_max: Maximum number of Reynolds numbers to plot. Useful if you ran a sweep with tons of Reynolds numbers.
        :param Cd_plot_max: Upper limit of Cd to plot [float]
        :return: self (makes plot)
        """

        self.has_xfoil_data()  # Ensure data is present.

        n_lines_max = min(n_lines_max, len(self.xfoil_data_2D["Re_indices"]))

        fig, ax = plt.subplots(1, 1, figsize=(7, 6), dpi=200)
        indices = np.array(
            np.round(np.linspace(0, len(self.xfoil_data_2D["Re_indices"]) - 1, n_lines_max)),
            dtype=int
        )
        indices_worth_plotting = [
            np.min(remove_nans(self.xfoil_data_2D["Cd"][index, :])) < Cd_plot_max
            for index in indices
        ]
        indices = indices[indices_worth_plotting]

        colors = plt.cm.rainbow(np.linspace(0, 1, len(indices)))[::-1]
        for i, Re in enumerate(self.xfoil_data_2D["Re_indices"][indices]):
            Cds = remove_nans(self.xfoil_data_2D["Cd"][indices[i], :])
            Cls = remove_nans(self.xfoil_data_2D["Cl"][indices[i], :])
            Cd_min = np.min(Cds)
            if Cd_min < Cd_plot_max:
                plt.plot(
                    Cds * 1e4,
                    Cls,
                    label="Re = %s" % eng_string(Re),
                    color=colors[i],
                )
        plt.xlim(0, Cd_plot_max * 1e4)
        plt.ylim(0, 2)
        plt.xlabel(r"$C_d \cdot 10^4$")
        plt.ylabel(r"$C_l$")
        plt.title("XFoil Polars for %s Airfoil" % self.name)
        plt.tight_layout()
        plt.legend()
        plt.show()

        return self

    def plot_xfoil_data_polar(self,
                              Res,  # type: list
                              Cd_plot_max=0.04,
                              repanel=False,
                              parallel=True,
                              max_iter=40,
                              verbose=True,
                              ):
        """
        Plots CL-CD polar for a single Reynolds number or a variety of Reynolds numbers.
        :param Res: Reynolds number to plot polars at. Either a single float or an iterable (list, 1D ndarray, etc.)
        :param Cd_plot_max: Upper limit of Cd to plot [float]
        :param cl_step: Cl increment for XFoil runs. Trades speed vs. plot resolution. [float]
        :param repanel: Should we repanel the airfoil within XFoil? [boolean]
        :param parallel: Should we run different Res in parallel? [boolean]
        :param max_iter: Maximum number of iterations for XFoil to run. [int]
        :param verbose: Should we print information as we run the sweeps? [boolean]
        :return: self (makes plot)
        """

        try:  # If it's not an iterable, make it one.
            Res[0]
        except TypeError:
            Res = [Res]

        fig, ax = plt.subplots(1, 1, figsize=(7, 6), dpi=200)
        colors = plt.cm.rainbow(np.linspace(0, 1, len(Res)))[::-1]

        def get_xfoil_data_at_Re(Re):

            import numpy as np  # needs to be imported here to support parallelization

            xfoil_data = self.xfoil_aseq(
                a_start=0,
                a_end=15,
                a_step=0.25,
                Re=Re,
                M=0,
                reset_bls=True,
                repanel=repanel,
                max_iter=max_iter,
                verbose=False,
            )
            Cd = remove_nans(xfoil_data["Cd"])
            Cl = remove_nans(xfoil_data["Cl"])
            return {"Cl": Cl, "Cd": Cd}

        if verbose:
            print("Running XFoil sweeps...")
            import time
            start_time = time.time()

        if not parallel:
            runs_data = [get_xfoil_data_at_Re(Re) for Re in Res]
        else:
            import multiprocess as mp
            pool = mp.Pool(mp.cpu_count())
            runs_data = pool.map(get_xfoil_data_at_Re, Res)
            pool.close()

        if verbose:
            run_time = time.time() - start_time
            print("XFoil Runtime: %.3f sec" % run_time)

        for i, Re in enumerate(Res):
            plt.plot(
                runs_data[i]["Cd"] * 1e4,
                runs_data[i]["Cl"],
                label="Re = %s" % eng_string(Re),
                color=colors[i],
            )
        plt.xlim(0, Cd_plot_max * 1e4)
        plt.ylim(0, 2)
        plt.xlabel(r"$C_d \cdot 10^4$")
        plt.ylabel(r"$C_l$")
        plt.title("XFoil Polars for %s Airfoil" % self.name)
        plt.tight_layout()
        plt.legend()
        plt.show()

        return self



def first_derivative(xs, ys, mode='centered'):
    if mode == 'forward':
        x = xs[:-1]
        y = ys[:-1]
        x_next = xs[1:]
        y_next = ys[1:]
        derivs = (y_next - y) / (x_next - x)
    elif mode == 'centered':
        x1 = xs[:-2]
        y1 = ys[:-2]
        x2 = xs[1:-1]
        y2 = ys[1:-1]
        x3 = xs[2:]
        y3 = ys[2:]
        derivs = (
                         x1 ** 2 * y2 + x2 ** 2 * y1 - x1 ** 2 * y3 + x3 ** 2 * y1
                         - x2 ** 2 * y3 - x3 ** 2 * y2 - 2 * x1 * x2 * y2 + 2 * x1 * x2 * y3
                         - 2 * x2 * x3 * y1 + 2 * x2 * x3 * y2
                 ) / (
                         (x1 - x2) * (x1 - x3) * (x2 - x3)
                 )
    else:
        raise Exception("Bad input for the ''mode'' argument.")
    return derivs


def second_derivative(xs, ys, mode='centered'):
    if mode == 'centered':
        x1 = xs[:-2]
        y1 = ys[:-2]
        x2 = xs[1:-1]
        y2 = ys[1:-1]
        x3 = xs[2:]
        y3 = ys[2:]
        derivs = -2 * (
                x1 * y2 - x2 * y1 - x1 * y3 + x3 * y1 + x2 * y3 - x3 * y2
        ) / (
                         (x1 - x2) * (x1 - x3) * (x2 - x3)
                 )
    else:
        raise Exception("Bad input for the ''mode'' argument.")
    return derivs


if __name__ == "__main__":
    import numpy as np

    x = np.linspace(0, 1, 11)**2
    y = x ** 2
    dydx = first_derivative(x, y, mode='forward')
    d2ydx2 = second_derivative(x, y)
    import matplotlib.pyplot as plt

    plt.plot(x, y, '.-')



from aerosandbox.geometry.common import *


class Airplane(AeroSandboxObject):
    """
    Definition for an airplane (or other vehicle/item to analyze).
    """

    def __init__(self,
                 name="Untitled",  # A sensible name for your airplane.
                 x_ref=0,  # Ref. point for moments; should be the center of gravity.
                 y_ref=0,  # Ref. point for moments; should be the center of gravity.
                 z_ref=0,  # Ref. point for moments; should be the center of gravity.
                 mass_props=None,  # An object of MassProps type; only needed for dynamic analysis
                 # If xyz_ref is not set, but mass_props is, the xyz_ref will be taken from the CG there.
                 wings=[],  # A list of Wing objects.
                 fuselages=[],  # A list of Fuselage objects.
                 s_ref=None,  # If not set, populates from first wing object.
                 c_ref=None,  # See above
                 b_ref=None,  # See above
                 ):
        self.name = name

        self.xyz_ref = cas.vertcat(x_ref, y_ref, z_ref)
        # if xyz_ref is None and mass_props is not None:
        #     self.xyz_ref = mass_props.get_cg()
        # else:
        #     self.xyz_ref = cas.MX(xyz_ref)
        # self.mass_props = mass_props
        self.wings = wings
        self.fuselages = fuselages

        if len(self.wings) > 0:  # If there is at least one wing
            self.set_ref_dims_from_wing(main_wing_index=0)
        if s_ref is not None:
            self.s_ref = s_ref
        if c_ref is not None:
            self.c_ref = c_ref
        if b_ref is not None:
            self.b_ref = b_ref

        # Check that everything was set right:
        assert self.name is not None
        assert self.xyz_ref is not None
        assert self.s_ref is not None
        assert self.c_ref is not None
        assert self.b_ref is not None

    def __repr__(self):
        return "Airplane %s (%i wings, %i fuselages)" % (
            self.name,
            len(self.wings),
            len(self.fuselages)
        )

    def set_ref_dims_from_wing(self,
                               main_wing_index=0
                               ):
        # Sets the reference dimensions of the airplane from measurements obtained from a specific wing.

        main_wing = self.wings[main_wing_index]

        self.s_ref = main_wing.area()
        self.b_ref = main_wing.span()
        self.c_ref = main_wing.mean_geometric_chord()

    def set_paneling_everywhere(self, n_chordwise_panels, n_spanwise_panels):
        # Sets the chordwise and spanwise paneling everywhere to a specified init_val.
        # Useful for quickly changing the fidelity of your simulation.

        for wing in self.wings:
            wing.chordwise_panels = n_chordwise_panels
            for xsec in wing.xsecs:
                xsec.spanwise_panels = n_spanwise_panels

    def set_spanwise_paneling_everywhere(self, n_spanwise_panels):
        # Sets the spanwise paneling everywhere to a specified value.
        # Useful for quickly changing the fidelity of your simulation.

        for wing in self.wings:
            for xsec in wing.xsecs:
                xsec.spanwise_panels = n_spanwise_panels

    def draw(self,
             show=True,  # type: bool
             colorscale="mint",  # type: str
             colorbar_title="Component ID",
             draw_quarter_chord=True,  # type: bool
             ):
        """
        Draws the airplane using a Plotly interface.
        :param show: Do you want to show the figure? [boolean]
        :param colorscale: Which colorscale do you want to use? ("viridis", "plasma", mint", etc.)
        :param draw_quarter_chord: Do you want to draw the quarter-chord? [boolean]
        :return: A plotly figure object [go.Figure]
        """
        fig = Figure3D()

        # Wings
        for wing_id in range(len(self.wings)):
            wing = self.wings[wing_id]  # type: Wing

            for xsec_id in range(len(wing.xsecs) - 1):
                xsec_1 = wing.xsecs[xsec_id]  # type: WingXSec
                xsec_2 = wing.xsecs[xsec_id + 1]  # type: WingXSec

                le_start = xsec_1.xyz_le + wing.xyz_le
                te_start = xsec_1.xyz_te() + wing.xyz_le
                le_end = xsec_2.xyz_le + wing.xyz_le
                te_end = xsec_2.xyz_te() + wing.xyz_le

                fig.add_quad(points=[
                    le_start,
                    le_end,
                    te_end,
                    te_start
                ],
                    intensity=wing_id,
                    mirror=wing.symmetric,
                )
                if draw_quarter_chord:
                    fig.add_line(  # draw the quarter-chord line
                        points=[
                            0.75 * le_start + 0.25 * te_start,
                            0.75 * le_end + 0.25 * te_end,
                        ],
                        mirror=wing.symmetric
                    )

        # Fuselages
        for fuse_id in range(len(self.fuselages)):
            fuse = self.fuselages[fuse_id]  # type: Fuselage

            for xsec_id in range(len(fuse.xsecs) - 1):
                xsec_1 = fuse.xsecs[xsec_id]  # type: FuselageXSec
                xsec_2 = fuse.xsecs[xsec_id + 1]  # type: FuselageXSec

                r1 = xsec_1.radius
                r2 = xsec_2.radius
                points_1 = np.zeros((fuse.circumferential_panels, 3))
                points_2 = np.zeros((fuse.circumferential_panels, 3))
                for point_index in range(fuse.circumferential_panels):
                    rot = angle_axis_rotation_matrix(
                        2 * cas.pi * point_index / fuse.circumferential_panels,
                        [1, 0, 0],
                        True
                    ).toarray()
                    points_1[point_index, :] = rot @ np.array([0, 0, r1])
                    points_2[point_index, :] = rot @ np.array([0, 0, r2])
                points_1 = points_1 + np.array(fuse.xyz_le).reshape(-1) + np.array(xsec_1.xyz_c).reshape(-1)
                points_2 = points_2 + np.array(fuse.xyz_le).reshape(-1) + np.array(xsec_2.xyz_c).reshape(-1)

                for point_index in range(fuse.circumferential_panels):

                    fig.add_quad(points=[
                        points_1[(point_index) % fuse.circumferential_panels, :],
                        points_1[(point_index + 1) % fuse.circumferential_panels, :],
                        points_2[(point_index + 1) % fuse.circumferential_panels, :],
                        points_2[(point_index) % fuse.circumferential_panels, :],
                    ],
                        intensity=fuse_id,
                        mirror=fuse.symmetric,
                    )

        return fig.draw(
            show=show,
            colorscale=colorscale,
            colorbar_title=colorbar_title,
        )

    def is_symmetric(self):
        """
        Returns a boolean describing whether the airplane is geometrically entirely symmetric across the XZ-plane.
        :return: [boolean]
        """
        for wing in self.wings:
            for xsec in wing.xsecs:
                if not (xsec.control_surface_type == "symmetric" or xsec.control_surface_deflection == 0):
                    return False
                if not wing.symmetric:
                    if not xsec.xyz_le[1] == 0:
                        return False
                    if not xsec.twist == 0:
                        if not (xsec.twist_axis[0] == 0 and xsec.twist_axis[2] == 0):
                            return False
                    if not xsec.airfoil.CL_function(0, 1e6, 0, 0) == 0:
                        return False
                    if not xsec.airfoil.Cm_function(0, 1e6, 0, 0) == 0:
                        return False

        return True

    def write_aswing(self, filepath=None):
        """
        Contributed by Brent Avery, Edited by Peter Sharpe. Work in progress.
        Writes a geometry file compatible with Mark Drela's ASWing.
        :param filepath: Filepath to write to. Should include ".asw" extension [string]
        :return: None
        """
        if filepath is None:
            filepath = "%s.asw" % self.name
        with open(filepath, "w+") as f:
            f.write('\n'.join(['#============',  # Name of Plane
                               'Name',
                               self.name,
                               'End']))
            f.write('\n'.join(['', '#============',  # Units that the analysis would be in, usually metric
                               'Units',
                               'L 0.3048 m',
                               'T 1.0  s',
                               'F 4.450 N',
                               'End']))
            f.write('\n'.join(['', '#============',  # Value of constants (cant imagine these changing too much)
                               'Constant',
                               '#  g     rho_0     a_0',
                               '   '.join([str(9.81), str(1.205), str(343.3)]),
                               'End']))
            f.write('\n'.join(['', '#============',  # Reference values (change automatically with input file)
                               'Reference',
                               '#   Sref    Cref    Bref',
                               '   '.join([str(self.s_ref), str(self.c_ref), str(self.b_ref)]),
                               'End']))

            '''
            Ok so 'ground' is a point on the plane that is constrained from translation or rotation.   
            Based on the documentation this is usually the 'frontest' part of the aircraft (why do I suck at words).   
            There is definitely a much better way to do this and I'm working on it but right now I'm just assuming   
            that it is the front part of the main wing, and just make that a constraint in AeroSandBox
            '''
            f.write('\n'.join(['', '#============',
                               'Ground',
                               '#  Nbeam  t',
                               '    '.join([' ', str(1), str(0)]),
                               'End']))

            f.write('\n'.join(['', '#============',
                               'Joint',
                               '#   Nbeam1   Nbeam2    t1     t2']))
            for onewing in range(1, len(self.wings)):
                wing = self.wings[onewing]
                if wing.name == "Horizontal Stabilizer":
                    xsecs = []
                    for xsec in wing.xsecs:
                        xsecs.append(xsec)
                    t = xsecs[0].y_le + wing.xyz_le[1]
                    coords = '       '.join(['    1', str(onewing + 1), str(t), '0'])
                    f.write('\n'.join(['', coords]))
                if wing.name == "Vertical Stabilizer":
                    wing2 = self.wings[np.ceil(onewing / 2)]
                    wing3 = self.wings[0]
                    xsecs = []
                    for xsec in wing.xsecs:
                        xsecs.append(xsec)
                    xsecs2 = []
                    for xsec2 in wing2.xsecs:
                        xsecs2.append(xsec2)
                    xsecs3 = []
                    for xsec3 in wing3.xsecs:
                        xsecs3.append(xsec3)
                    t = xsecs2[0].y_le + wing.xyz_le[1]
                    t2 = 1 + (xsecs2[0].z_le + wing2.xyz_le[2]) - (xsecs[0].z_le + wing.xyz_le[2])
                    coords = '       '.join(['    1', str(onewing + 1), str(t), str(t2)])
                    f.write('\n'.join(['', coords]))

            for fuse in range(len(self.fuselages)):
                onefuse = self.fuselages[fuse]
                wing = self.wings[0]
                xsecs = []
                xsecs.append(onefuse.xsecs[0])
                xsecs.append(onefuse.xsecs[-1])
                xsecs2 = []
                for xsec2 in wing.xsecs:
                    xsecs2.append(xsec2)
                t = xsecs[0].y_c + onefuse.xyz_le[1]
                t2 = ((wing.xyz_le[0] + xsecs2[0].x_le) + (wing.xyz_le[0] + xsecs2[-1].x_le)) / 2
                coords = '      '.join(['    1', str(onewing + fuse + 2), str(t), str(t2)])
                f.write('\n'.join(['', coords]))

            corr_stab = {}

            for fuse in range(len(self.fuselages)):
                corr_stab.update({fuse + len(self.wings) + 1: [fuse + 1, fuse + 1 + np.floor(len(self.wings) / 2)]})

            for fuse in range(len(self.fuselages)):
                onefuse = self.fuselages[fuse]
                xsecs = []
                xsecs.append(onefuse.xsecs[0])
                xsecs.append(onefuse.xsecs[-1])
                horiz = self.wings[corr_stab[fuse + len(self.wings) + 1][0]]
                xsecs2 = []
                for xsec2 in horiz.xsecs:
                    xsecs2.append(xsec2)
                vert = self.wings[corr_stab[fuse + len(self.wings) + 1][1]]
                xsecs3 = []
                for xsec3 in vert.xsecs:
                    xsecs3.append(xsec3)
                t = xsecs2[0].x_le + horiz.xyz_le[0]
                t2 = 0
                t3 = xsecs3[0].x_le + vert.xyz_le[0]
                t4 = 1 + (xsecs2[0].z_le + horiz.xyz_le[2]) - (xsecs3[0].z_le + vert.xyz_le[2])
                coords = '     '.join(
                    ['', str(fuse + len(self.wings) + 1), str(corr_stab[fuse + len(self.wings) + 1][0] + 1), str(t),
                     str(t2)])
                coords2 = '     '.join(
                    ['', str(fuse + len(self.wings) + 1), str(corr_stab[fuse + len(self.wings) + 1][1] + 1), str(t3),
                     str(t4)])
                f.write('\n'.join(['', coords, coords2]))
            f.write('\n'.join(['', 'End']))
            '''
            The juicy stuff! This part of the code iterates over each wing and then subiterates (is that a word?) over 
            each wing's cross section. Along the way it collects information on chord length, angle, and coordinates of
            all the leading edges. It then writes all this info in a way that ASWing likes
            '''
            for onewing in range(len(self.wings)):
                wing = self.wings[onewing]
                if wing.name == "Main Wing":
                    xsecs = []
                    for xsec in wing.xsecs:
                        xsecs.append(xsec)
                    chordalfa = []
                    coords = []
                    '''
                    This part is hard to explain but basically I defined t (the beamwise axis)
                    as the axis that the beam changes most along. This can be generalized
                    but I'm not entirely sure how
                    '''
                    max_le = {abs(xsecs[-1].x_le - xsecs[0].x_le): 'sec.x_le', \
                              abs(xsecs[-1].y_le - xsecs[0].y_le): 'sec.y_le', \
                              abs(xsecs[-1].z_le - xsecs[0].z_le): 'sec.z_le'}
                    for sec in xsecs:
                        if max_le.get(max(max_le)) == 'sec.x_le':
                            t = sec.x_le
                        elif max_le.get(max(max_le)) == 'sec.y_le':
                            t = sec.y_le
                        elif max_le.get(max(max_le)) == 'sec.z_le':
                            t = sec.z_le
                        chordalfa.append('    '.join([str(t), str(sec.chord), str(sec.twist)]))
                        coords.append(
                            '    '.join([str(t), str(sec.x_le + wing.xyz_le[0]), str(sec.y_le + wing.xyz_le[1]),
                                         str(sec.z_le + wing.xyz_le[2])]))
                    f.write('\n'.join(['', '#============',
                                       ' '.join(['Beam', str(onewing + 1)]),
                                       wing.name,
                                       't    chord    twist',
                                       '\n'.join(chordalfa),
                                       '#',
                                       't    x    y    z',
                                       '\n'.join(coords),
                                       'End']))
                elif wing.name == "Horizontal Stabilizer":
                    xsecs = []
                    for xsec in wing.xsecs:
                        xsecs.append(xsec)
                    chordalfa = []
                    coords = []
                    '''
                    This part is hard to explain but basically I defined t (the beamwise axis)
                    as the axis that the beam changes most along. This can be generalized
                    but I'm not entirely sure how
                    '''
                    max_le = {abs(xsecs[-1].x_le - xsecs[0].x_le): 'sec.x_le', \
                              abs(xsecs[-1].y_le - xsecs[0].y_le): 'sec.y_le', \
                              abs(xsecs[-1].z_le - xsecs[0].z_le): 'sec.z_le'}
                    for sec in xsecs:
                        if max_le.get(max(max_le)) == 'sec.x_le':
                            t = sec.x_le
                        elif max_le.get(max(max_le)) == 'sec.y_le':
                            t = sec.y_le
                        elif max_le.get(max(max_le)) == 'sec.z_le':
                            t = sec.z_le
                        chordalfa.append('    '.join([str(t), str(sec.chord), str(sec.twist), str(0.07)]))
                        coords.append(
                            '    '.join([str(t), str(sec.x_le + wing.xyz_le[0]), str(sec.y_le + wing.xyz_le[1]),
                                         str(sec.z_le + wing.xyz_le[2])]))
                    f.write('\n'.join(['', '#============',
                                       ' '.join(['Beam', str(onewing + 1)]),
                                       wing.name,
                                       't    chord    twist dCLdF1',
                                       '\n'.join(chordalfa),
                                       '#',
                                       't    x    y    z',
                                       '\n'.join(coords),
                                       'End']))

                elif wing.name == "Vertical Stabilizer":
                    xsecs = []
                    for xsec in wing.xsecs:
                        xsecs.append(xsec)
                    chordalfa = []
                    coords = []
                    '''
                    This part is hard to explain but basically I defined t (the beamwise axis)
                    as the axis that the beam changes most along. This can be generalized
                    but I'm not entirely sure how
                    '''
                    max_le = {abs(xsecs[-1].x_le - xsecs[0].x_le): 'sec.x_le', \
                              abs(xsecs[-1].y_le - xsecs[0].y_le): 'sec.y_le', \
                              abs(xsecs[-1].z_le - xsecs[0].z_le): 'sec.z_le'}
                    for sec in xsecs:
                        if max_le.get(max(max_le)) == 'sec.x_le':
                            t = sec.x_le + 1
                        elif max_le.get(max(max_le)) == 'sec.y_le':
                            t = sec.y_le + 1
                        elif max_le.get(max(max_le)) == 'sec.z_le':
                            t = sec.z_le + 1
                        chordalfa.append('    '.join([str(t), str(sec.chord), str(sec.twist)]))
                        coords.append(
                            '    '.join([str(t), str(sec.x_le + wing.xyz_le[0]), str(sec.y_le + wing.xyz_le[1]),
                                         str(sec.z_le + wing.xyz_le[2])]))
                    f.write('\n'.join(['', '#============',
                                       ' '.join(['Beam', str(onewing + 1)]),
                                       wing.name,
                                       '  t    chord    twist',
                                       '\n'.join(chordalfa),
                                       '#',
                                       't    x    y    z',
                                       '\n'.join(coords),
                                       'End']))

            for fuse in range(len(self.fuselages)):
                onefuse = self.fuselages[fuse]
                xsecs = []
                xsecs.append(onefuse.xsecs[0])
                xsecs.append(onefuse.xsecs[-1])
                coords = []
                max_c = {abs(xsecs[1].x_c - xsecs[0].x_c): 'sec.x_c', \
                         abs(xsecs[1].y_c - xsecs[0].y_c): 'sec.y_c', \
                         abs(xsecs[1].z_c - xsecs[0].z_c): 'sec.z_c'}
                for sec in xsecs:
                    if max_c.get(max(max_c)) == 'sec.x_c':
                        t = sec.x_c
                    elif max_c.get(max(max_c)) == 'sec.y_c':
                        t = sec.y_c
                    elif max_c.get(max(max_c)) == 'sec.z_c':
                        t = sec.z_c
                    coords.append(
                        '    '.join([str(t), str(sec.x_c + onefuse.xyz_le[0]), str(sec.y_c + onefuse.xyz_le[1]),
                                     str(sec.z_c + onefuse.xyz_le[2])]))
                f.write('\n'.join(['', '#============',
                                   '  '.join(['Beam', str(onewing + fuse + 2)]),
                                   onefuse.name,
                                   't    x    y    z',
                                   '\n'.join(coords),
                                   'End']))

    def approximate_longitudinal_center_of_pressure(self):
        """
        Returns the approximate location of the center of pressure. Given as the area-weighted quarter chord of the wing.
        :return: [x, y, z] of the approximate center of pressure
        """
        areas = [wing.area(type="projected") for wing in self.wings]
        COPs = [wing.approximate_center_of_pressure() for wing in self.wings]

        total_area = cas.sum1(cas.vertcat(*areas))

        COP = cas.sum1(cas.vertcat(*[
            COPs[i]*areas[i]/total_area
            for i in range(len(self.wings))
        ]))

        return COP


import numpy as np
import casadi as cas
from aerosandbox.tools.miscellaneous import stdout_redirected


def fit(
        model,  # type: callable
        x_data,  # type: dict
        y_data,  # type: np.ndarray
        param_guesses,  # type: dict
        param_bounds=None,  # type: dict
        weights=None,  # type: np.ndarray
        verbose=True,  # type: bool
        scale_problem=True,  # type: bool
        put_residuals_in_logspace=False,  # type: bool
):
    """
    Fits a model to data through least-squares minimization.
    :param model: A callable with syntax f(x, p) where:
            x is a dict of dependent variables. Same format as x_data [dict of 1D ndarrays of length n].
            p is a dict of parameters. Same format as param_guesses [dict of scalars].
        Model should use CasADi functions for differentiability.
    :param x_data: a dict of dependent variables. Same format as model's x. [dict of 1D ndarrays of length n]
    :param y_data: independent variable. [1D ndarray of length n]
    :param param_guesses: a dict of fit parameters. Same format as model's p. Keys are parameter names, values are initial guesses. [dict of scalars]
    :param param_bounds: Optional: a dict of bounds on fit parameters.
        Keys are parameter names, values are a tuple of (min, max).
        May contain only a subset of param_guesses if desired.
        Use None to represent one-sided constraints (i.e. (None, 5)).
        [dict of tuples]
    :param weights: Optional: weights for data points. If not supplied, weights are assumed to be uniform.
        Weights are automatically normalized. [1D ndarray of length n]
    :param verbose: Whether or not to print information about parameters and goodness of fit.
    :param scale_problem: Whether or not to attempt to scale variables, constraints, and objective for more robust solve. [boolean]
    :param put_residuals_in_logspace: Whether to optimize using the logarithmic error as opposed to the absolute error (useful for minimizing percent error).
        Note: If any model outputs or data are negative, this will fail!
    :return: Optimal fit parameters [dict]
    """
    opti = cas.Opti()

    # Handle weighting
    if weights is None:
        weights = cas.GenDM_ones(y_data.shape[0])
    weights /= cas.sum1(weights)

    def fit_param(initial_guess, lower_bound=None, upper_bound=None):
        """
        Helper function to create a fit variable
        :param initial_guess:
        :param lower_bound:
        :param upper_bound:
        :return:
        """
        if scale_problem and np.abs(initial_guess) > 1e-8:
            var = initial_guess * opti.variable()  # scale variables
        else:
            var = opti.variable()
        opti.set_initial(var, initial_guess)
        if lower_bound is not None:
            lower_bound_abs = np.abs(lower_bound)
            if scale_problem and lower_bound_abs > 1e-8:
                opti.subject_to(var / lower_bound_abs > lower_bound / lower_bound_abs)
            else:
                opti.subject_to(var > lower_bound)
        if upper_bound is not None:
            upper_bound_abs = np.abs(upper_bound)
            if scale_problem and upper_bound_abs > 1e-8:
                opti.subject_to(var / upper_bound_abs < upper_bound / upper_bound_abs)
            else:
                opti.subject_to(var < upper_bound)
        return var

    if param_bounds is None:
        params = {
            k: fit_param(param_guesses[k])
            for k in param_guesses
        }
    else:
        params = {
            k: fit_param(param_guesses[k]) if k not in param_bounds else
            fit_param(param_guesses[k], param_bounds[k][0], param_bounds[k][1])
            for k in param_guesses
        }

    if scale_problem:
        y_model_initial = model(x_data, param_guesses)
        if not put_residuals_in_logspace:
            residuals_initial = y_model_initial - y_data
        else:
            residuals_initial = cas.log10(y_model_initial) - cas.log10(y_data)
        SSE_initial = cas.sum1(weights * residuals_initial ** 2)

        y_model = model(x_data, params)
        if not put_residuals_in_logspace:
            residuals = y_model - y_data
        else:
            residuals = cas.log10(y_model) - cas.log10(y_data)
        SSE = cas.sum1(weights * residuals ** 2)
        opti.minimize(SSE / SSE_initial)

    else:
        y_model = model(x_data, params)
        if not put_residuals_in_logspace:
            residuals = y_model - y_data
        else:
            residuals = cas.log10(y_model) - cas.log10(y_data)
        SSE = cas.sum1(weights * residuals ** 2)
        opti.minimize(SSE)

    # Solve
    p_opts = {}
    s_opts = {}
    s_opts["max_iter"] = 3e3  # If you need to interrupt, just use ctrl+c
    # s_opts["mu_strategy"] = "adaptive"
    opti.solver('ipopt', p_opts, s_opts)
    opti.solver('ipopt')
    if verbose:
        sol = opti.solve()
    else:
        with stdout_redirected():
            sol = opti.solve()

    params_solved = {}
    for k in params:
        try:
            params_solved[k] = sol.value(params[k])
        except:
            params_solved[k] = np.NaN

    # printing
    if verbose:
        # Print parameters
        print("\nFit Parameters:")
        if len(params_solved) <= 20:
            [print("\t%s = %f" % (k, v)) for k, v in params_solved.items()]
        else:
            print("\t%i parameters solved for." % len(params_solved))
        print("\nGoodness of Fit:")

        # Print RMS error
        weighted_RMS_error = sol.value(cas.sqrt(cas.sum1(
            weights * residuals ** 2
        )))
        print("\tWeighted RMS error: %f" % weighted_RMS_error)

        # Print R^2
        y_data_mean = cas.sum1(y_data) / y_data.shape[0]
        SS_tot = cas.sum1(weights * (y_data - y_data_mean) ** 2)
        SS_res = cas.sum1(weights * (y_data - y_model) ** 2)
        R_squared = sol.value(1 - SS_res / SS_tot)
        print("\tR^2: %f" % R_squared)

    return params_solved



## Snippets of code that I'm not using anymore, but it's good to keep them on hand.

# class Panel:
#     def __init__(self,
#                  vertices=None,  # Nx3 np array, each row is a vector. Just used for drawing panel
#                  colocation_point=None,  # 1x3 np array
#                  normal_direction=None,  # 1x3 np array, nonzero
#                  influencing_objects=[],  # List of Vortexes and Sources
#                  ):
#         self.vertices = np.array(vertices)
#         self.colocation_point = np.array(colocation_point)
#         self.normal_direction = np.array(normal_direction)
#         self.influencing_objects = influencing_objects
#
#         assert (np.shape(self.vertices)[0] >= 3)
#         assert (np.shape(self.vertices)[1] == 3)
#
#     def centroid(self):
#         return np.mean(self.vertices, axis=0)
#
#     def area(self):
#         centroid = self.centroid()
#         area = 0
#         for i in range(len(self.vertices) - 1):
#             area += 0.5 * np.linalg.norm(
#                 np.cross(
#                     self.vertices[i, :] - centroid,
#                     self.vertices[i + 1, :] - centroid
#                 )
#             )
#         return area
#
#     def set_colocation_point_at_centroid(self):
#         centroid = np.mean(self.vertices, axis=0)
#         self.colocation_point = centroid
#
#     def add_ring_vortex(self):
#         pass
#
#     def calculate_influence(self, point):
#         influence = np.zeros([3])
#         for object in self.influencing_objects:
#             influence += object.calculate_unit_influence(point)
#         return influence
#
#     def draw(self,
#              show=True,
#              fig_to_plot_on=None,
#              ax_to_plot_on=None,
#              draw_colocation_point=True,
#              shading_color=None,
#              ):
#
#         # Setup
#         if fig_to_plot_on == None or ax_to_plot_on == None:
#             fig, ax = fig3d()
#             fig.set_size_inches(12, 9)
#         else:
#             fig = fig_to_plot_on
#             ax = ax_to_plot_on
#
#         # Plot vertices
#         if not (self.vertices == None).all():
#             quad = Poly3DCollection([self.vertices],
#                                     cmap=mpl.cm.get_cmap('viridis'),
#                                     )
#             quad.set_edgecolor(None)
#             quad.set_facecolor(shading_color)
#             ax.add_collection3d(quad)
#
#             # verts_to_draw = np.vstack((self.vertices, self.vertices[0, :]))
#             # x = verts_to_draw[:, 0]
#             # y = verts_to_draw[:, 1]
#             # z = verts_to_draw[:, 2]
#             # ax.plot(x, y, z, color='#be00cc', linestyle='-', linewidth=0.4)
#
#         # Plot colocation point
#         if draw_colocation_point and not (self.colocation_point == None).all():
#             x = self.colocation_point[0]
#             y = self.colocation_point[1]
#             z = self.colocation_point[2]
#             ax.scatter(x, y, z, color=(0, 0, 0), marker='.', s=0.5)
#
#         if show:
#             plt.show()
#
#
# class HorseshoeVortex:
#     # As coded, can only have two points not at infinity (3-leg horseshoe vortex)
#     # Wake assumed to trail to infinity in the x-direction.
#     def __init__(self,
#                  vertices=None,  # 2x3 np array, left point first, then right.
#                  strength=0,
#                  ):
#         self.vertices = np.array(vertices)
#         self.strength = np.array(strength)
#
#         assert (self.vertices.shape == (2, 3))
#
#     def calculate_unit_influence(self, point):
#         # Calculates the velocity induced at a point per unit vortex strength
#         # Taken from Drela's Flight Vehicle Aerodynamics, pg. 132
#
#         a = point - self.vertices[0, :]
#         b = point - self.vertices[1, :]
#         norm_a = np.linalg.norm(a)
#         norm_b = np.linalg.norm(b)
#         x_hat = np.array([1, 0, 0])
#
#         influence = 1 / (4 * np.pi) * (
#                 (np.cross(a, b) / (norm_a * norm_b + a @ b)) * (1 / norm_a + 1 / norm_b) +
#                 (np.cross(a, x_hat) / (norm_a - a @ x_hat)) / norm_a -
#                 (np.cross(b, x_hat) / (norm_b - b @ x_hat)) / norm_b
#         )
#
#         return influence
#
#
# class Source:
#     # A (3D) point source/sink.
#     def __init__(self,
#                  vertex=None,
#                  strength=0,
#                  ):
#         self.vertex = np.array(vertex)
#         self.strength = np.array(strength)
#
#         assert (self.vertices.shape == (3))
#
#     def calculate_unit_influence(self, point):
#         r = self.vertices - point
#         norm_r = np.linalg.norm(r)
#
#         influence = 1 / (4 * np.pi * norm_r ** 2)
#
#         return influence
